import time
from pymycobot.common import ProtocolCode
from pymycobot.error import calibration_parameters
import math
import threading


class ultraArmP340:
    def __init__(self, port, baudrate=115200, timeout=0.1, debug=False):
        """Args:
        port     : port string
        baudrate : baud rate string, default '115200'
        timeout  : default 0.1
        debug    : whether show debug info, default: False
        """
        import serial

        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = True
        self._serial_port.dtr = True
        self._serial_port.open()
        self.debug = debug
        self.calibration_parameters = calibration_parameters
        self.lock = threading.Lock()
        time.sleep(1)

    def _respone(self):
        """Send data to the robot"""
        time.sleep(0.02)
        data = None
        if self._serial_port.inWaiting() > 0:
            data = self._serial_port.read(self._serial_port.inWaiting())
            queue_size = self._get_queue_size()
            if data != None:
                if 40 <= queue_size <= 90:
                    data = str(data.decode())
                    if self.debug:
                        print(data)
                elif 0 <= queue_size < 40:
                    data = str(data.decode())
                    if self.debug:
                        print(data)
                else:
                    qs = 0
                    while True:
                        try:
                            time.sleep(2)
                            qs = self._get_queue_size()
                            if self.debug:
                                print("respone_size1:", qs)
                            if qs < 40:
                                time.sleep(1)
                                qsize = self._get_queue_size()
                                print("respone_size2:", qsize)
                                break

                        except Exception as e:
                            print(e)

    def _request(self, flag=""):
        """Get data from the robot"""
        time.sleep(0.02)
        data = None
        count = 0
        while count < 5:
            if self._serial_port.inWaiting() > 0:
                data = self._serial_port.read(self._serial_port.inWaiting())
                data = str(data.decode())
                if self.debug:
                    print(
                        "\nreceived data:\n%s**********************\n" % data
                    )
                if data.find("ERROR: COMMAND NOT RECOGNIZED") > -1:
                    flag = None

                if flag == "angle":
                    a = data[data.find("ANGLES") :]
                    astart = a.find("[")
                    aend = a.find("]")
                    if a != None and a != "":
                        try:
                            all = list(
                                map(float, a[astart + 1 : aend].split(","))
                            )
                            return all
                        except Exception:
                            print(
                                "received angles is not completed! Retry receive..."
                            )
                            count += 1
                            continue

                elif flag == "coord":
                    c = data[data.find("COORDS") :]
                    cstart = c.find("[")
                    cend = c.find("]")
                    if c != None and c != "":
                        try:
                            all = list(
                                map(float, c[cstart + 1 : cend].split(","))
                            )
                            return all
                        except Exception:
                            print(
                                "received coords is not completed! Retry receive..."
                            )
                            count += 1
                            continue

                elif flag == "endstop":
                    e = data[data.find("ENDSTOP") :]
                    eend = e.find("\n")
                    if e != None and e != "":
                        try:
                            all = e[:eend]
                            return all
                        except Exception:
                            print(
                                "received coords is not completed! Retry receive..."
                            )
                            count += 1
                            continue

                elif flag == "size":
                    try:
                        q = data[data.find("QUEUE SIZE") :]
                        qstart = q.find("[")
                        qend = q.find("]")
                        if q != None and q != "":
                            qsize = int(q[qstart + 1 : qend])
                            return qsize
                    except Exception as e:
                        count += 1
                        continue
                elif flag == "isStop":
                    if "Moving end" in data:
                        return 1
                    else:
                        return 0
                elif flag == "gripper":
                    header = "GRIPPERANGLE["
                    read = data.find(header) + len(header)
                    # print(data[read:])
                    end = data[read:].find("]")
                    return data[read : read + end]
                elif flag == "system":
                    header = "ReadSYS["
                    read = data.find(header) + len(header)
                    end = data[read:].find("]")
                    return int(data[read : read + end])
                elif flag == "system_version":
                    start_idx = data.find("GetSystemVersion[")
                    if start_idx != -1:
                        version_str = data[
                            start_idx + len("GetSystemVersion[") :
                        ]
                        version_str = version_str.split("]")[0]
                        version = float(version_str.strip())
                        return version
                    else:
                        return None
                elif flag == "modify_version":
                    start_idx = data.find("GetModifyVersion[")
                    if start_idx != -1:
                        version_str = data[
                            start_idx + len("GetModifyVersion[") :
                        ]
                        version_str = version_str.split("]")[0]
                        version = int(version_str.strip())
                        return version
                    else:
                        return None
                elif flag == None:
                    return 0

    def _debug(self, data):
        """Whether show info."""
        if self.debug:
            print("\n***** Debug Info *****\nsend command: %s" % data)

    def _get_queue_size(self):
        """Get real queue_size from the robot."""
        command = "M120" + ProtocolCode.END
        self._serial_port.write(command.encode())
        self._serial_port.flush()
        self._debug(command)
        return self._request("size")

    def release_all_servos(self):
        """Relax all joints."""
        with self.lock:
            command = ProtocolCode.RELEASE_SERVOS + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def power_on(self):
        """Lock all joints."""
        with self.lock:
            command = ProtocolCode.LOCK_SERVOS + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def go_zero(self):
        """Back to zero."""
        with self.lock:
            command = ProtocolCode.BACK_ZERO + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            time.sleep(0.1)
            data = None
            while True:
                if self._serial_port.inWaiting() > 0:
                    data = self._serial_port.read(
                        self._serial_port.inWaiting()
                    )
                    if data != None:
                        data = str(data.decode())
                        if data.find("ok") > 0:
                            if self.debug:
                                print(data)
                            break

    def sleep(self, wait_time):
        """Set wait time

        Args:
            wait_time (int): wait time 1 ~ 65535
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, wait_time=wait_time
        )
        with self.lock:
            command = (
                ProtocolCode.SLEEP_TIME
                + " S"
                + str(wait_time)
                + ProtocolCode.END
            )
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def get_angles_info(self):
        """Get the current joint angle information."""
        with self.lock:
            command = ProtocolCode.GET_CURRENT_ANGLE_INFO + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._request("angle")

    def get_radians_info(self):
        """Get the current joint radian information."""
        with self.lock:
            while True:
                angles = self.get_angles_info()
                if angles != None:
                    break
            radians = []
            for i in angles:
                radians.append(round(i * (math.pi / 180), 3))
            return radians

    def get_coords_info(self):
        """Get current Cartesian coordinate information."""
        with self.lock:
            command = ProtocolCode.GET_CURRENT_COORD_INFO + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._request("coord")

    def get_switch_state(self):
        """Get the current state of all home switches."""
        with self.lock:
            command = ProtocolCode.GET_BACK_ZERO_STATUS + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._request("endstop")

    def set_init_pose(self, pose_coords):
        """Set the current coords to zero.

        Args:
            pose_coords (list[float]): [X, Y, Z, E] coordinate values.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, pose_coords=pose_coords
        )
        with self.lock:
            command = ProtocolCode.SET_JOINT
            command += " X" + str(pose_coords[0])
            command += " Y" + str(pose_coords[1])
            command += " Z" + str(pose_coords[2])
            if len(pose_coords) == 4:
                if pose_coords[3] is not None:
                    command += " E" + str(pose_coords[3])
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_coords(self, coords, speed):
        """Set all joints coords.

        Args:
            coords: a list of coords value(List[float]).
                x : -360 ~ 365.55 mm
                y : -365.55 ~ 365.55 mm
                z : -140 ~ 130 mm
            speed : (int) 1-200 mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coords=coords, speed=speed
        )

        with self.lock:
            command = ProtocolCode.COORDS_SET
            command += " X" + str(coords[0])
            command += " Y" + str(coords[1])
            command += " Z" + str(coords[2])
            if len(coords) == 4:
                if coords[3] is not None:
                    command += " E" + str(coords[3])
            command += " F" + str(speed)
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_coord(self, coord_id, coord, speed):
        """Set single coordinate.

        Args:
            coord_id (str): 'X', 'Y', 'Z', 'E'
            coord (float): coordinate value
            speed (int): movement speed (1-200 mm/s)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__,
            coord_id=coord_id,
            coord=coord,
            speed=speed,
        )
        with self.lock:
            command = ProtocolCode.COORDS_SET
            command += f" {coord_id}{coord}"
            command += f" F{speed}"
            command += ProtocolCode.END

            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_mode(self, mode):
        """Set coord mode.

        Args:
            mode:
                0 - Absolute Cartesian mode

                1 - Relative Cartesian mode
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, mode=mode
        )
        with self.lock:
            if mode == 0:
                command = ProtocolCode.ABS_CARTESIAN + ProtocolCode.END
            else:
                command = ProtocolCode.REL_CARTESIAN + ProtocolCode.END

            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_speed_mode(self, speed_mode):
        """Set speed mode.

        Args:
            speed_mode:
                0 - Constant speed mode

                2 - Acceleration / deceleration mode
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, speed_mode=speed_mode
        )
        with self.lock:
            command = "M16 S" + str(speed_mode) + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_pwm(self, p_value):
        """PWM control.

        Args:
            p_value (int) : Duty cycle 0 ~ 255; 128 means 50%
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, p_value=p_value
        )
        with self.lock:
            command = (
                ProtocolCode.SET_PWM + "p" + str(p_value) + ProtocolCode.END
            )
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_gpio_state(self, state):
        """Set gpio state.

        Args:
            state (int):
                1 - open
                0 - close
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, state=state
        )
        with self.lock:
            if state:
                command = ProtocolCode.GPIO_CLOSE + ProtocolCode.END
            else:
                command = ProtocolCode.GPIO_ON + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_gripper_zero(self):
        """Set gripper zero."""
        with self.lock:
            command = ProtocolCode.GRIPPER_ZERO + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_gripper_state(self, gripper_value, gripper_speed):
        """Set gripper angle.

        Args:
            gripper_value (int): 0 - 100

            gripper_speed: 1 - 1500
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__,
            gripper_value=gripper_value,
            gripper_speed=gripper_speed,
        )
        with self.lock:
            command = (
                ProtocolCode.GIRPPER_OPEN
                + "A"
                + str(gripper_value)
                + "F"
                + str(gripper_speed)
                + ProtocolCode.END
            )
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_gripper_release(self):
        """Set gripper release."""
        with self.lock:
            command = ProtocolCode.GIRPPER_RELEASE + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_fan_state(self, state):
        """Set fan state.

        Args:
            state (int):
                0 - close
                1 - open
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, state=state
        )
        with self.lock:
            if state:
                command = ProtocolCode.FAN_ON + ProtocolCode.END
            else:
                command = ProtocolCode.FAN_CLOSE + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_angle(self, joint_id, angle, speed):
        """Set single angle.

        Args:
            joint_id (int) : joint (1/2/3/4)
            angle :
                1 : -150° ~ +170°
                2 : -20° ~ 90°
                3 : -5° ~ 110°
                4 : -179° ~ + 179°
            speed : (int) 1-200 mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__,
            joint_id=joint_id,
            angle=angle,
            speed=speed,
        )
        with self.lock:
            command = ProtocolCode.SET_ANGLE
            command += " J" + str(joint_id)
            command += " A" + str(angle)
            command += " F" + str(speed)
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_angles(self, angles, speed):
        """Set all joints angles.

        Args:
            angles: A list of angles value(List[float]).
            speed : (int) 1-200 mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, angles=angles, speed=speed
        )
        with self.lock:
            command = ProtocolCode.SET_ANGLES
            command += " X" + str(angles[0])
            command += " Y" + str(angles[1])
            command += " Z" + str(angles[2])
            if len(angles) == 4:
                command += " E" + str(angles[3])
            command += " F" + str(speed)
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_radians(self, degrees, speed):
        """Set all joints radians

        Args:
            degrees: A list of radians value(List[float]).
                    1 : -2.6179 ~ 2.9670
                    2 : -0.3490 ~ 1.5707
                    3 : -0.0872 ~ 1.9198
                    4 : -3.1241 ~ + 3.1241
            speed : (int) 1-200 mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, radians=degrees, speed=speed
        )
        with self.lock:
            degrees = [
                round(degree * (180 / math.pi), 2) for degree in degrees
            ]
            self.set_angles(degrees, speed)

    def set_jog_angle(self, joint_id, direction, speed):
        """Start jog movement with angle

        Args:
            joint_id : joint(1/2/3)

            direction :
                1 : Negative motion
                0 : Positive motion
            speed : (int) 1-200 mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__,
            joint_id=joint_id,
            direction=direction,
            speed=speed,
        )
        with self.lock:
            command = ProtocolCode.SET_JOG_ANGLE
            command += " J" + str(joint_id)
            command += " D" + str(direction)
            command += " F" + str(speed)
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_jog_coord(self, axis_id, direction, speed):
        """Start jog movement with coord

        Args:
            axis_id(int) : axis 1-X, 2-Y, 3-Z

            direction:
                1 : Negative motion
                0 : Positive motion
            speed : (int) 1-200 mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__,
            axis_id=axis_id,
            direction=direction,
            speed=speed,
        )
        with self.lock:
            command = ProtocolCode.JOG_COORD_
            command += " J" + str(axis_id)
            command += " D" + str(direction)
            command += " F" + str(speed)
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def set_jog_stop(self):
        """Stop jog movement"""
        with self.lock:
            command = ProtocolCode.SET_JOG_STOP
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            self._respone()

    def play_gcode_file(self, filename):
        """Play the imported track file

        Args:
            filename (str): Path to a G-code file (.gcode or .nc or .ngc)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, filename=filename
        )
        X = []
        try:
            with open(filename) as f:
                lines = f.readlines()
                for line in lines:
                    line = line.strip("\n")
                    X.append(line)
        except Exception:
            print("There is no such file!")

        begin, end = 0, len(X)

        self.set_speed_mode(0)  # Constant speed mode

        with self.lock:
            for i in range(begin, end):
                command = X[i] + ProtocolCode.END
                self._serial_port.write(command.encode())
                self._serial_port.flush()
                time.sleep(0.02)

                self._debug(command)

                queue_size = self._request("size")
                if 0 <= queue_size <= 90:
                    try:
                        if self.debug:
                            print("queue_size1: %s \n" % queue_size)
                    except Exception as e:
                        print(e)
                else:
                    qs = 0
                    while True:
                        try:
                            qs = self._get_queue_size()
                            if self.debug:
                                print("queue_size2: %s \n" % qs)
                            if qs <= 40:
                                begin = i
                                break
                        except Exception:
                            print("Retry receive...")
                command = ""

        self.set_speed_mode(2)  # Acceleration / deceleration mode

    def close(self):
        with self.lock:
            self._serial_port.close()

    def open(self):
        with self.lock:
            self._serial_port.open()

    def is_moving_end(self):
        with self.lock:
            """Get the current state of all home switches."""
            command = ProtocolCode.IS_MOVING_END + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._request("isStop")

    def sync(self):
        while self.is_moving_end() != 1:
            pass

    def get_gripper_angle(self):
        """Get gripper angle"""
        command = ProtocolCode.GET_GRIPPER_ANGLE + ProtocolCode.END
        self._serial_port.write(command.encode())
        self._serial_port.flush()
        self._debug(command)
        return self._request("gripper")

    def set_system_value(self, id, address, value, mode=None):
        """Set system parameters

        Args:
            id (int): 4 or 7
            address (int): 7 ~ 69
            value (int):
            mode (int): 1 or 2, can be empty, default mode is 1
                1 - setting range is 0-255, address 21 (P value) can be used
                2 - setting value range is 0-65535, address 56 (setting position) can be used
        """
        if mode:
            self.calibration_parameters(
                class_name=self.__class__.__name__,
                id=id,
                address=address,
                system_value=value,
                system_mode=mode,
            )
            command = (
                ProtocolCode.SET_SYSTEM_VALUE
                + f" X{id} "
                + f"Y{address} "
                + f"Z{value} "
                + f"P{mode} "
                + ProtocolCode.END
            )
        else:
            self.calibration_parameters(
                class_name=self.__class__.__name__,
                id=id,
                address=address,
                system_value=value,
            )
            command = (
                ProtocolCode.SET_SYSTEM_VALUE
                + f" X{id} "
                + f"Y{address} "
                + f"Z{value} "
                + ProtocolCode.END
            )
        self._serial_port.write(command.encode())
        self._serial_port.flush()
        self._debug(command)

    def get_system_value(self, id, address, mode=None):
        """Get system parameters

        Args:
            id (int): 4 or 7
            address (int): 0 ~ 69
            mode (int): 1 or 2, can be empty, default mode is 1
                1 - read range is 0-255, address 21 (P value) can be used
                2 - read value range is 0-65535, address 56 (read position) can be used

        Returns:
            int: 0 ~ 65535
        """
        if mode:
            self.calibration_parameters(
                class_name=self.__class__.__name__,
                id=id,
                get_address=address,
                system_mode=mode,
            )
            command = (
                ProtocolCode.GET_SYSTEM_VALUE
                + f" J{id} "
                + f"S{address} "
                + f"P{mode} "
                + ProtocolCode.END
            )
        else:
            self.calibration_parameters(
                class_name=self.__class__.__name__, id=id, get_address=address
            )
            command = (
                ProtocolCode.GET_SYSTEM_VALUE
                + f" J{id} "
                + f"S{address} "
                + ProtocolCode.END
            )
        self._serial_port.write(command.encode())
        self._serial_port.flush()
        self._debug(command)
        return self._request("system")

    def get_system_version(self):
        """Get system firmware version

        Returns:
            (float) Firmware version
        """
        command = ProtocolCode.GET_SYSTEM_VERSION + ProtocolCode.END
        self._serial_port.write(command.encode())
        self._serial_port.flush()
        self._debug(command)
        return self._request("system_version")

    def get_modify_version(self):
        """Get firmware modify version

        Returns:
            (int) modify version
        """
        command = ProtocolCode.GET_MODIFY_VERSION + ProtocolCode.END
        self._serial_port.write(command.encode())
        self._serial_port.flush()
        self._debug(command)
        return self._request("modify_version")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import asyncio
from .myagvpro import MyAGVProCommandProtocolApi, ProtocolCode, PLAINTEXT_REPLY_PROTOCOL_CODE


class MyAGVProCommandApi(MyAGVProCommandProtocolApi):

    async def _match_protocol_data(self, genre, timeout=0.1):
        for _ in range(5):
            reply_data = await self.read()

            if len(reply_data) == 0:
                continue

            self._save_buffer_data(data=reply_data, to_local=True)

            if genre in PLAINTEXT_REPLY_PROTOCOL_CODE:
                if not reply_data.startswith(b"AGVPro:"):
                    continue
            else:
                if reply_data[3] != genre.value:
                    continue

            self.log.info(f" read: {' '.join(f'{x:02x}' for x in reply_data)}")
            break
        else:
            reply_data = None
        return reply_data

    async def _merge(self, genre, *args):
        with self._mutex:
            timeout = self.genre_timeout_table.get(genre, 0.1)
            if not ProtocolCode.GET_AUTO_REPORT_MESSAGE.equal(genre):
                real_command = self._combination(genre, args)
                self.log.info(f"write: {' '.join(f'{x:02x}' for x in real_command)}")
                await self.write(real_command)

            reply_data = await self._match_protocol_data(genre, timeout)
            decode_respond = self._instruction_decoding(genre, reply_data)
            return self._parsing_data(genre, decode_respond)

    async def get_system_version(self):
        """Obtain the major firmware version number

        Returns:
            float: version
        """
        return await self._merge(ProtocolCode.GET_SYSTEM_VERSION)

    async def get_modify_version(self):
        """Obtain the minor firmware version number

        Returns:
            int: version
        """
        return await self._merge(ProtocolCode.GET_MODIFY_VERSION)

    async def get_robot_status(self):
        """Obtain the machine status, and support the acquisition in the case of power failure

        Returns:
            tuple: (list[int] | int, float)
                0 - list[int] | int: machine status，Normally 0
                    0 - (int)Emergency stop status
                    1 - (int)Power status
                    2 - (int)Front bumper strip
                    3 - (int)Rear bumper strip
                    4 - (int)Motor No. 1 connection status
                    5 - (int)Motor No. 2 connection status
                    6 - (int)Motor No. 3 connection status
                    7 - (int)Motor No. 4 connection status
        """
        return await self._merge(ProtocolCode.GET_ROBOT_STATUS)

    async def power_on(self):
        """Turn on the robot

        Returns:
            int: Power-on result
                1 - Success
                2 - Emergency stop triggered
                3 - The battery is too low
                4 - CAN initialization is abnormal
                5 - Motor initialization exception
        """
        return await self._merge(ProtocolCode.POWER_ON)

    async def power_on_only(self):
        """Turn on the robot, but not start the control program

        Returns:
            int: Power-on result only, 1: Success, 0: Failed
        """
        return await self._merge(ProtocolCode.POWER_ON_ONLY)

    async def power_off(self):
        """Turn off the robot

        Returns:
            int: Power-off result only, 1: Success, 0: Failed
        """
        return await self._merge(ProtocolCode.POWER_OFF)

    async def is_power_on(self):
        """Check if the robot is powered on

        Returns:
            int: power state, 1: Power-on, 0: Power-off
        """
        return await self._merge(ProtocolCode.GET_POWER_STATE)

    async def move_forward(self, speed):
        """Pan the robot forward

        Args:
            speed(float): 0.01 ~ 1.50m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.50:
            raise ValueError("Speed must be between 0.01 and 1.50")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return await self._merge(ProtocolCode.AGV_MOTION_CONTROL, [int(speed * 100 * 1), 0x00])

    async def move_backward(self, speed):
        """Pan the robot backward

        Args:
            speed(float): 0.01 ~ 1.50m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.50:
            raise ValueError("Speed must be between 0.01 and 1.50")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return await self._merge(ProtocolCode.AGV_MOTION_CONTROL, [int(speed * 100 * -1)])

    async def move_left_lateral(self, speed):
        """Pan the robot left

        Args:
            speed(float): 0.01 ~ 1.00 m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.00:
            raise ValueError("Speed must be between 0.01 and 1.00")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return await self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, int(speed * 100 * -1)])

    async def move_right_lateral(self, speed):
        """Pan the robot right

        Args:
            speed(float): 0.01 ~ 1.00m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.00:
            raise ValueError("Speed must be between 0.00 and 1.00")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return await self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, int(speed * 100 * 1)])

    async def turn_left(self, speed):
        """Rotate to the left

        Args:
            speed:

        Returns:
            int: 1: Success, 0: Failed
        """
        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return await self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, 0x00, int(speed * 100 * -1), 0x00])

    async def turn_right(self, speed):
        """Rotate to the right

        Args:
            speed:

        Returns:
            int: 1: Success, 0: Failed
        """
        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return await self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, 0x00, int(speed * 100 * 1), 0x00])

    async def stop(self):
        """Stop moving

        Returns:
            int: 1: Success, 0: Failed
        """
        return await self._merge(ProtocolCode.AGV_STOP_MOVING)

    async def set_auto_report_state(self, state):
        """Set the auto-report state

        Args:
            state(int): 0: Close, 1: Open

        Returns:
            int: 1: Success, 0: Failed
        """

        if state not in (0, 1):
            raise ValueError("State must be 0 or 1")

        return await self._merge(ProtocolCode.SET_AUTO_REPORT_STATE, state)

    async def get_auto_report_state(self):
        """Get the auto-report state

        Returns:
            int: 0: Close, 1: Open
        """
        return await self._merge(ProtocolCode.GET_AUTO_REPORT_STATE)

    async def get_auto_report_message(self):
        """Get the auto-report message
        Returns:
            list[int | list[int] | float]:
             0 - (float)rx
             1 - (float)ry
             2 - (float)rw
             3 - (list[int])Machine status
             4 - (list[int])Motor information
             5 - (float)Battery voltage
             6 - (int)Motor enable status 0: Enabled, 1: Disabled
        """
        return await self._merge(ProtocolCode.GET_AUTO_REPORT_MESSAGE)

    async def set_motor_enable(self, motor_id, state):
        """Enable or disable the motor

        Args:
            motor_id(int):
                1: Left upper motor
                2: Right upper motor
                3: Left lower motor
                4: Right lower motor
                254: All motors
            state(bool):
                0: Disable
                1: Enable
        Returns:
            int: 1: Success, 0: Failed
        """
        if motor_id not in (1, 2, 3, 4, 254):
            raise ValueError("Motor id must be 0 or 1")

        if state not in (0, 1):
            raise ValueError("State must be 0 or 1")

        return await self._merge(ProtocolCode.SET_MOTOR_ENABLED, motor_id, state)

    async def get_motor_status(self):
        """Get the motor status

        Returns:
            list[int]: Motor status
                0: normal
                any: error code
        """
        return await self._merge(ProtocolCode.GET_MOTOR_STATUS)

    async def get_motor_temps(self):
        """Get the motor temperature

        Returns:
            list[float]: Motor temperature
        """
        return await self._merge(ProtocolCode.GET_MOTOR_TEMPERATURE)

    async def get_motor_speeds(self):
        """Get the motor speeds

        Returns:
            list[float]: Motor speeds
        """
        return await self._merge(ProtocolCode.GET_MOTOR_SPEEDS)

    async def get_motor_torques(self):
        """Get the motor torques

        Returns:
            list[float]: Motor torques
        """
        return await self._merge(ProtocolCode.GET_MOTOR_TORQUES)

    async def get_motor_enable_status(self):
        """Get the motor enabled status
        Returns:
            list[int]: Motor enabled status
                0: Disable
                1: Enable
        """
        return await self._merge(ProtocolCode.GET_MOTOR_ENABLE_STATUS)

    async def get_motor_loss_count(self):
        """Get the motor loss count

        Returns:
            list[int]: Motor loss count
        """
        return await self._merge(ProtocolCode.GET_MOTOR_LOSS_COUNT)

    async def set_communication_state(self, state):
        """Set the communication state

        Args:
            state(int):
                0: Serial communication (default)
                1: Socket communication
                2: Bluetooth communication (Write the MAC address to the file and the endpoint, and then return to the state)

        Returns:
            int: 1: Success, 0: Failed
        """
        if state not in (0, 1, 2):
            raise ValueError("State must be 0, 1 or 2")
        self._communication_mode = state
        return await self._merge(ProtocolCode.SET_COMMUNICATION_MODE, state)

    async def get_communication_state(self):
        """Get the communication state

        Returns:
            int: communication state
                0: Serial communication,
                1: Socket communication,
                2: Bluetooth communication
        """
        return await self._merge(ProtocolCode.GET_COMMUNICATION_MODE)

    async def set_led_color(self, position, color, brightness=255):
        """Set the LED color

        Args:
            position(int):
                0: Left LED
                1: Right LED
            color(tuple(int, int, int)): RGB color
            brightness(int): 0 - 255 (default: 255)

        Returns:
            int: 1: Success, 0: Failed
        """
        if position not in (0, 1):
            raise ValueError("Position must be 0 or 1")

        if not 0 <= brightness <= 255:
            raise ValueError("Brightness must be between 0 and 255")

        if len(color) != 3:
            raise ValueError("Color must be a tuple of 3 values")

        if any(map(lambda c: not 0 <= c <= 255, color)):
            raise ValueError("Color must be between 0 and 255")

        return await self._merge(ProtocolCode.SET_LED_COLOR, position, brightness, *color)

    async def set_led_mode(self, mode):
        """Set the LED mode

        Args:
            mode(int):
                0: Battery level(default)
                1: DIY

        Returns:
            int: 1: Success, 0: Failed
        """
        if mode not in (0, 1):
            raise ValueError("Mode must be 0 or 1")
        return await self._merge(ProtocolCode.SET_LED_MODE, mode)

    async def set_pin_output(self, pin, state):
        """Set the output IO

        Args:
            pin(int): 1 - 6
            state(int): 0: Low, 1: High

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 1 <= pin <= 6:
            raise ValueError("Pin must be between 1 and 6")

        if state not in (0, 1):
            raise ValueError("State must be 0 or 1")

        return await self._merge(ProtocolCode.SET_OUTPUT_IO, pin, state)

    async def get_pin_input(self, pin):
        """Get the input IO

        Args:
            pin(int): 1 - 6

        Returns:
            int: 0: Low, 1: High, -1: There is no such pin
        """
        if not 1 <= pin <= 6:
            raise ValueError("Pin must be between 1 and 6")
        return await self._merge(ProtocolCode.GET_INPUT_IO, pin)

    def get_estop_state(self):
        """Get the emergency stop state

        Returns:
            int: 0: Release, 1: Press
        """
        return self._merge(ProtocolCode.GET_INPUT_IO, 254)

    async def get_wifi_account(self):
        """Get the wi-fi account

        Returns:
            tuple(str, str): wi-fi account, wi-fi password
        """
        return await self._merge(ProtocolCode.GET_WIFI_ACCOUNT)

    async def get_wifi_ip_port(self):
        """Get the wi-fi ip and port
        Returns:
            tuple(str, int): wi-fi ip, wi-fi port
        """
        return await self._merge(ProtocolCode.GET_WIFI_IP_PORT)

    async def get_bluetooth_uuid(self):
        """Get the bluetooth uuid

        Returns:
            tuple(str, str, str): bluetooth name, service uuid, characteristic uuid
        """

        return await self._merge(ProtocolCode.GET_BLUETOOTH_UUID)

    async def get_bluetooth_address(self):
        """Get the bluetooth MAC address

        Returns:
            str: bluetooth MAC address
        """

        return await self._merge(ProtocolCode.GET_BLUETOOTH_ADDRESS)

    async def set_handle_control_state(self, state):
        """Set the handle control switch status
        Args:
            state(int): 0: Disable, 1: Enable

        Returns:
            int: 1: Success, 0: Failed
        """
        if state not in (0, 1):
            raise ValueError("state must be 0 or 1")
        return await self._merge(ProtocolCode.SET_HANDLE_CONTROL_STATE, state)

    async def get_handle_control_state(self):
        """Get the handle control switch status

        Returns:
            int: 0: Disable, 1: Enable
        """
        return await self._merge(ProtocolCode.GET_HANDLE_CONTROL_STATE)


class MyAGVProBluetooth(MyAGVProCommandApi):
    def __init__(self, address, service_uuid, char_uuid, debug=False, save_serial_log=False):
        super().__init__(debug=debug, save_serial_log=save_serial_log)
        from bleak import BleakClient
        self._bluetooth = BleakClient(address, timeout=10)
        self._address = address
        self._char_uuid = char_uuid
        self._service_uuid = service_uuid
        self._serial_filename = 'agvpro_bluetooth_serial.log'
        self._communication_mode = 2
        self._queue = asyncio.Queue()

    # Async Context managers
    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.disconnect()

    async def read(self, size=1):
        if not self._queue.empty():
            return self._queue.get_nowait()
        return await self._bluetooth.read_gatt_char(self._char_uuid, size=size)

    async def write(self, data):
        return await self._bluetooth.write_gatt_char(self._char_uuid, data)

    async def connect(self):
        await self._bluetooth.connect()
        await self._bluetooth.start_notify(self._char_uuid, self._handle_notification)

    async def _handle_notification(self, _, data):
        await self._queue.put(data)

    async def disconnect(self):
        await self._bluetooth.stop_notify(self._char_uuid)
        await self._bluetooth.disconnect()

    async def is_connected(self):
        return self._bluetooth.is_connected

    @classmethod
    async def connect_by_name(cls, name, service_uuid, char_uuid, debug=False, save_serial_log=False):
        from bleak import BleakScanner
        device = await BleakScanner.find_device_by_name(name=name)
        if device is None:
            return None

        return cls(
            address=device,
            service_uuid=service_uuid,
            char_uuid=char_uuid,
            debug=debug,
            save_serial_log=save_serial_log
        )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from __future__ import division

import functools
import socket
import time
import threading
import serial

from pymycobot.generate import DataProcessor
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters


def setup_serial_port(port, baudrate, timeout=None):
    serial_api = serial.Serial()
    serial_api.port = port
    serial_api.baudrate = baudrate
    serial_api.timeout = timeout
    serial_api.rts = False
    serial_api.open()
    return serial_api


def setup_socket_connect(host, port, timeout):
    socket_api = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_api.settimeout(timeout)
    socket_api.connect((host, port))
    return socket_api


class GPIOProtocolCode:
    SETUP_GPIO_MODE = 0xAA
    SETUP_GPIO_STATE = 0xAB
    SET_GPIO_OUTPUT = 0xAC
    GET_GPIO_INPUT = 0xAD


class MyCobot280RDKX5Api(DataProcessor):

    def __init__(self, debug=False, thread_lock=True):
        super(MyCobot280RDKX5Api, self).__init__(debug)
        self.calibration_parameters = functools.partial(self.parametric_verification, class_name="MyCobot280")
        self.thread_lock = thread_lock
        self.lock = threading.Lock()

    @classmethod
    def parametric_verification(cls, **kwargs):
        kwargs["class_name"] = "MyCobot280RDK-X5"
        return calibration_parameters(**kwargs)

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply, _ = super(MyCobot280RDKX5Api, self)._mesg(genre, *args, **kwargs)
        if self.thread_lock:
            with self.lock:
                return self._res(real_command, has_reply, genre)
        else:
            return self._res(real_command, has_reply, genre)

    def _res(self, real_command, has_reply, genre):
        if genre == ProtocolCode.SET_SSID_PWD or genre == ProtocolCode.GET_SSID_PWD:
            self._write(self._flatten(real_command))
            data = self._read(genre)
        else:
            try_count = 0
            while try_count < 3:
                self._write(self._flatten(real_command))
                data = self._read(genre)
                if data is not None and data != b'':
                    break
                try_count += 1
            else:
                return -1
        if genre == ProtocolCode.SET_SSID_PWD:
            return 1

        res = self._process_received(data, genre)

        if res is None:
            return -1
        if genre in [ProtocolCode.SET_BASIC_OUTPUT]:
            return 1
        if res is not None and isinstance(res, list) and len(res) == 1 and genre not in [
            ProtocolCode.GET_BASIC_VERSION, ProtocolCode.GET_JOINT_MIN_ANGLE, ProtocolCode.GET_JOINT_MAX_ANGLE,
            ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION
        ]:
            return res[0]

        if genre in [
            ProtocolCode.IS_POWER_ON,
            ProtocolCode.IS_CONTROLLER_CONNECTED,
            ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
            ProtocolCode.IS_IN_POSITION,
            ProtocolCode.IS_MOVING,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.IS_ALL_SERVO_ENABLE,
            ProtocolCode.GET_SERVO_DATA,
            ProtocolCode.GET_DIGITAL_INPUT,
            ProtocolCode.GET_GRIPPER_VALUE,
            ProtocolCode.IS_GRIPPER_MOVING,
            ProtocolCode.GET_SPEED,
            ProtocolCode.GET_ENCODER,
            ProtocolCode.GET_BASIC_INPUT,
            ProtocolCode.GET_TOF_DISTANCE,
            ProtocolCode.GET_END_TYPE,
            ProtocolCode.GET_MOVEMENT_TYPE,
            ProtocolCode.GET_REFERENCE_FRAME,
            ProtocolCode.GET_FRESH_MODE,
            ProtocolCode.GET_GRIPPER_MODE,
            ProtocolCode.GET_ERROR_INFO,
            ProtocolCode.GET_COMMUNICATE_MODE,
            ProtocolCode.SET_COMMUNICATE_MODE,
            ProtocolCode.SetHTSGripperTorque,
            ProtocolCode.GetHTSGripperTorque,
            ProtocolCode.GetGripperProtectCurrent,
            ProtocolCode.InitGripper,
            ProtocolCode.SET_FOUR_PIECES_ZERO
        ]:
            return self._process_single(res)
        elif genre in [ProtocolCode.GET_ANGLES, ProtocolCode.SOLVE_INV_KINEMATICS]:
            return [self._int2angle(angle) for angle in res]
        elif genre in [ProtocolCode.GET_COORDS, ProtocolCode.GET_TOOL_REFERENCE, ProtocolCode.GET_WORLD_REFERENCE]:
            if res:
                r = []
                for idx in range(3):
                    r.append(self._int2coord(res[idx]))
                for idx in range(3, 6):
                    r.append(self._int2angle(res[idx]))
                return r
            else:
                return res
        elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
            return [self._int2coord(angle) for angle in res]
        elif genre in [ProtocolCode.GET_JOINT_MAX_ANGLE, ProtocolCode.GET_JOINT_MIN_ANGLE]:
            return self._int2coord(res[0])
        elif genre in [
            ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION
        ]:
            return self._int2coord(self._process_single(res))
        elif genre in [ProtocolCode.GET_REBOOT_COUNT]:
            return self._process_high_low_bytes(res)
        elif genre in (ProtocolCode.GET_ANGLES_COORDS, ):
            r = []
            for index, el in enumerate(res):
                if index < 6:
                    r.append(self._int2angle(el))
                elif index < 9:
                    r.append(self._int2coord(el))
                elif index < 12:
                    r.append(self._int2angle(el))
                else:
                    r.append(el)
            return r
        else:
            return res

    def wait(self, t):
        time.sleep(t)
        return self


class MyCobot280RDKX5CommandGenerator(MyCobot280RDKX5Api):

    # Overall Status
    def get_modify_version(self):
        """get modify version"""
        return self._mesg(ProtocolCode.ROBOT_VERSION)

    def get_system_version(self):
        """get system version"""
        return self._mesg(ProtocolCode.SOFTWARE_VERSION)

    def clear_queue(self):
        """Clear the command queue"""
        return self._mesg(ProtocolCode.CLEAR_COMMAND_QUEUE)

    def get_queue_length(self):
        """Get current queue length"""
        return self._mesg(ProtocolCode.GET_COMMAND_QUEUE)

    def async_or_sync(self):
        """Set the command execution mode
        Return:
            0: Asynchronous mode
            1: Synchronous mode
        """
        return self._mesg(ProtocolCode.CHECK_ASYNC_OR_SYNC)

    def get_error_information(self):
        """Obtaining robot error information

        Return:
            0: No error message.
            1 ~ 6: The corresponding joint exceeds the limit position.
            16 ~ 19: Collision protection.
            32: Kinematics inverse solution has no solution.
            33 ~ 34: Linear motion has no adjacent solution.
        """
        return self._mesg(ProtocolCode.GET_ERROR_INFO, has_reply=True)

    def clear_error_information(self):
        """Clear robot error message"""
        return self._mesg(ProtocolCode.CLEAR_ERROR_INFO, has_reply=True)

    def set_monitor_state(self, monitor_state):
        """
        Set the monitoring state
        Args:
            monitor_state: 0 - Disable monitoring
                   1 - Enable monitoring
        """
        self.calibration_parameters(monitor_state=monitor_state)
        return self._mesg(ProtocolCode.SET_MONITOR_STATE, monitor_state)

    def get_monitor_state(self):
        """
        Get the monitoring state
        """
        return self._mesg(ProtocolCode.GET_MONITOR_STATE, has_reply=True)

    def get_reboot_count(self):
        """Get the number of times the robot has been restarted"""
        return self._mesg(ProtocolCode.GET_REBOOT_COUNT, has_reply=True)

    def power_on(self):
        """Open communication with Atom."""
        return self._mesg(ProtocolCode.POWER_ON)

    def power_off(self):
        """Close communication with Atom."""
        return self._mesg(ProtocolCode.POWER_OFF)

    def is_power_on(self):
        """Adjust robot arm status

        Return:
            1 - power on
            0 - power off
            -1 - error data
        """
        return self._mesg(ProtocolCode.IS_POWER_ON, has_reply=True)

    def release_all_servos(self):
        """Relax all joints"""
        return self._mesg(ProtocolCode.RELEASE_ALL_SERVOS)

    def focus_all_servos(self):
        """Damping all joints"""
        return self._mesg(ProtocolCode.FOCUS_ALL_SERVOS)

    def read_next_error(self):
        """Robot Error Detection

        Return:
            list len 6.
            0 : No abnormality
            1 : Communication disconnected
            2 : Unstable communication
            3 : Servo abnormality
        """
        return self._mesg(ProtocolCode.READ_NEXT_ERROR, has_reply=True)

    def set_fresh_mode(self, fresh_mode):
        """Set command refresh mode

        Args:
            fresh_mode: int.
                1 - Always execute the latest command first.
                0 - Execute instructions sequentially in the form of a queue.
        """
        self.calibration_parameters(fresh_mode=fresh_mode)
        return self._mesg(ProtocolCode.SET_FRESH_MODE, fresh_mode)

    def get_fresh_mode(self):
        """Query sports mode"""
        return self._mesg(ProtocolCode.GET_FRESH_MODE, has_reply=True)

    def set_vision_mode(self, vision_mode):
        """Set the visual tracking mode to limit the posture flipping of send_coords in refresh mode.
        (Only for visual tracking function)

        Args:
            vision_mode: 0/1; 0 - close; 1 - open
        """
        self.calibration_parameters(vision_mode=vision_mode)
        return self._mesg(ProtocolCode.SET_VISION_MODE, vision_mode)

    # Motion interface
    def get_angles(self):
        """ Get the angle of all joints.

        Return:
            list: A float list of all angle.
        """
        return self._mesg(ProtocolCode.GET_ANGLES, has_reply=True)

    def send_angle(self, joint_id, angle, speed, _async=False):
        """Send one angle of joint to robot arm.

        Args:
            joint_id : Joint id(genre.Angle) int 1-6.
            angle : angle value(float).
            speed : (int) 1 ~ 100
            _async: (default False) Whether to execute asynchronously, Currently invalid
        """
        self.calibration_parameters(joint_id=joint_id, angle=angle, speed=speed)
        return self._mesg(ProtocolCode.SEND_ANGLE, joint_id, [self._angle2int(angle)], speed, has_reply=True, _async=_async)

    def send_angles(self, angles, speed, _async=False):
        """Send the angles of all joints to robot arm.

        Args:
            angles: a list of angle values(List[float]). len 6.
            speed : (int) 1 ~ 100
            _async: (default False) Whether to execute asynchronously, Currently invalid
        """
        self.calibration_parameters(angles=angles, speed=speed)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.SEND_ANGLES, angles, speed, has_reply=True, _async=_async)

    def get_coords(self):
        """Get the coords from robot arm, coordinate system based on base.

        Return:
            list : A float list of coord .[x, y, z, rx, ry, rz]
        """
        return self._mesg(ProtocolCode.GET_COORDS, has_reply=True)

    def angles_to_coords(self, angles):
        """ Convert angles to coordinates

        Args:
            angles : A float list of all angle.

        Return:
            list: A float list of all coordinates.
        """
        self.calibration_parameters(angles=angles)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.GET_COORDS, angles, has_reply=True)

    def send_coord(self, coord_id, coord, speed, _async=False):
        """Send one coord to robot arm.

        Args:
            coord_id(int) : coord id(genre.Coord) int 1-6.
            coord(float) : coord value, mm
            speed(int) : 1 ~ 100
            _async: (default False) Whether to execute asynchronously, Currently invalid
        """
        self.calibration_parameters(coord_id=coord_id, coord=coord, speed=speed)
        value = self._coord2int(coord) if coord_id <= 3 else self._angle2int(coord)
        return self._mesg(ProtocolCode.SEND_COORD, coord_id, [value], speed, has_reply=True, _async=_async)

    def send_coords(self, coords, speed, coord_mode=None, _async=False):
        """Send all coords to robot arm.

        Args:
            coords: a list of coords value(List[float]).[x(mm), y, z, rx(angle), ry, rz]\n
            speed : (int) 1 ~ 100
            coord_mode : (int) 0 - angluar, 1 - linear
            _async: (default False) Whether to execute asynchronously, Currently invalid
        """
        self.calibration_parameters(coords=coords, speed=speed, coord_mode=coord_mode)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        if coord_mode is not None:
            return self._mesg(ProtocolCode.SEND_COORDS, coord_list, speed, coord_mode, _async=_async)
        else:
            return self._mesg(ProtocolCode.SEND_COORDS, coord_list, speed, has_reply=True, _async=_async)

    def sync_send_angles(self, degrees, speed, timeout=15):
        """Send the angle in synchronous state and return when the target point is reached

        Args:
            degrees: a list of degree values(List[float]), length 6.
            speed: (int) 0 ~ 100
            timeout: default 15s.
        """
        t = time.time()
        self.send_angles(degrees, speed)
        while time.time() - t < timeout:
            f = self.is_in_position(degrees, 0)
            if f == 1:
                break
            time.sleep(0.1)
        return 1

    def sync_send_coords(self, coords, speed, mode=0, timeout=15):
        """Send the coord in synchronous state and return when the target point is reached

        Args:
            coords: a list of coord values(List[float])
            speed: (int) 0 ~ 100
            mode: (int): 0 - angular（default）, 1 - linear
            timeout: default 15s.
        """
        t = time.time()
        self.send_coords(coords, speed, mode)
        while time.time() - t < timeout:
            if self.is_in_position(coords, 1) == 1:
                break
            time.sleep(0.1)
        return 1

    def pause(self):
        """Pause movement"""
        return self._mesg(ProtocolCode.PAUSE)

    def is_paused(self):
        """Judge whether the manipulator pauses or not.

        Return:
            1 - paused
            0 - not paused
            -1 - error
        """
        return self._mesg(ProtocolCode.IS_PAUSED, has_reply=True)

    def resume(self):
        """Recovery movement"""
        return self._mesg(ProtocolCode.RESUME)

    def stop(self):
        """Stop moving"""
        return self._mesg(ProtocolCode.STOP)

    def is_in_position(self, data, mode=0):
        """Judge whether in the position.

        Args:
            data: A data list, angles or coords.len 6.
            mode: 1 - coords, 0 - angles

        Return:
            1 - True\n
            0 - False\n
            -1 - Error
        """
        if mode == 1:
            self.calibration_parameters(coords=data)
            data_list = []
            for idx in range(3):
                data_list.append(self._coord2int(data[idx]))
            for idx in range(3, 6):
                data_list.append(self._angle2int(data[idx]))
        elif mode == 0:
            self.calibration_parameters(angles=data)
            data_list = [self._angle2int(i) for i in data]
        else:
            raise Exception("mode is not right, please input 0 or 1")

        return self._mesg(ProtocolCode.IS_IN_POSITION, data_list, id, has_reply=True)

    def is_moving(self):
        """Detect if the robot is moving

        Return:
            0 - not moving
            1 - is moving
            -1 - error data
        """
        return self._mesg(ProtocolCode.IS_MOVING, has_reply=True)

    # JOG mode and operation
    def jog_angle(self, joint_id, direction, speed):
        """Jog control angle.

        Args:
            joint_id: int 1-6.
            direction: 0 - decrease, 1 - increase
            speed: int (0 - 100)
        """
        self.calibration_parameters(joint_id=joint_id, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_ANGLE, joint_id, direction, speed)

    def jog_rpy(self, end_direction, direction, speed):
        """Rotate the end around a fixed axis in the base coordinate system

        Args:
            end_direction (int):  Roll, Pitch, Yaw (1-3)
            direction (int): 1 - forward rotation, 0 - reverse rotation
            speed (int): 1 ~ 100
        """
        self.calibration_parameters(end_direction=end_direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_ABSOLUTE, end_direction, direction, speed)

    def jog_coord(self, coord_id, direction, speed):
        """Jog control coord.

        Args:
            coord_id: int 1-6
            direction: 0 - decrease, 1 - increase
            speed: int (1 - 100)
        """
        self.calibration_parameters(coord_id=coord_id, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_COORD, coord_id, direction, speed)

    def jog_increment_angle(self, joint_id, increment, speed):
        """ angle step mode

        Args:
            joint_id: int 1-6.
            increment: Angle increment value
            speed: int (1 - 100)
        """
        self.calibration_parameters(id=joint_id, speed=speed, angle=increment)
        return self._mesg(ProtocolCode.JOG_INCREMENT, joint_id, [self._angle2int(increment)], speed)

    def jog_increment_coord(self, coord_id, increment, speed):
        """coord step mode

        Args:
            coord_id: coord id 1 - 6.
            increment: Coord increment value
            speed: int (1 - 100)
        """
        self.calibration_parameters(coord_id=coord_id, speed=speed, coord=increment)
        value = self._coord2int(increment) if coord_id <= 3 else self._angle2int(increment)
        return self._mesg(ProtocolCode.JOG_INCREMENT_COORD, coord_id, [value], speed)

    def set_HTS_gripper_torque(self, gripper_torque):
        """Set new adaptive gripper torque

        Args:
            gripper_torque (int): 150 ~ 980

        Return:
            0: Set failed
            1: Set successful
        """
        self.calibration_parameters(gripper_torque=gripper_torque)
        return self._mesg(ProtocolCode.SetHTSGripperTorque, [gripper_torque], has_reply=True)

    def get_HTS_gripper_torque(self):
        """Get gripper torque

        Returns:
            int: 150 ~ 980
        """
        return self._mesg(ProtocolCode.GetHTSGripperTorque, has_reply=True)

    def get_gripper_protect_current(self):
        """Get the gripper protection current

        Returns:
            int: 1 ~ 500
        """
        return self._mesg(ProtocolCode.GetGripperProtectCurrent, has_reply=True)

    def init_gripper(self):
        """Initialize gripper

        Returns:
            int: 0 or 1 (1 - success)
        """
        return self._mesg(ProtocolCode.InitGripper, has_reply=True)

    def set_gripper_protect_current(self, protect_current):
        """Set the gripper protection current

        Args:
            protect_current (int): 1 ~ 500
        """
        self.calibration_parameters(protect_current=protect_current)

        return self._mesg(ProtocolCode.SetGripperProtectCurrent, [protect_current])

    def set_encoder(self, servo_id, encoder, speed):
        """Set a single joint rotation to the specified potential value.

        Args:
            servo_id: int  1 - 7
            encoder: The value of the set encoder.
            speed : 1 - 100
        """
        self.calibration_parameters(servo_id=servo_id, encoder=encoder, speed=speed)
        return self._mesg(ProtocolCode.SET_ENCODER, servo_id, [encoder], speed)

    def get_encoder(self, servo_id):
        """Obtain the specified joint potential value.

        Args:
            servo_id: (int) 1 - 6
                for gripper: Joint id 7
        """
        self.calibration_parameters(servo_id=servo_id)
        return self._mesg(ProtocolCode.GET_ENCODER, servo_id, has_reply=True)

    def set_encoders(self, encoders, speed):
        """Set the six joints of the manipulator to execute synchronously to the specified position.

        Args:
            encoders: A encoder list. len 6.
            speed: speed 1 ~ 100
        """
        self.calibration_parameters(encoders=encoders, speed=speed)
        return self._mesg(ProtocolCode.SET_ENCODERS, encoders, speed)

    def set_encoders_drag(self, encoders, speeds):
        """Send all encoders and speeds

        Args:
            encoders: encoders list.
            speeds: Obtained by the get_servo_speeds() method
        """
        self.calibration_parameters(encoders=encoders, drag_speeds=speeds)
        return self._mesg(ProtocolCode.SET_ENCODERS_DRAG, encoders, speeds)

    def get_joint_min_angle(self, joint_id):
        """Gets the minimum movement angle of the specified joint

        Args:
            joint_id: 1 - 6

        Return:
            angle value(float)
        """
        self.calibration_parameters(joint_id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MIN_ANGLE, joint_id, has_reply=True)

    def get_joint_max_angle(self, joint_id):
        """Gets the maximum movement angle of the specified joint

        Args:
            joint_id: 1 - 6

        Return:
            angle value(float)
        """
        self.calibration_parameters(joint_id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MAX_ANGLE, joint_id, has_reply=True)

    def set_joint_max(self, joint_id, angle):
        """Set the joint maximum angle

        Args:
            joint_id: int.
                Joint id 1 - 6
                for gripper: Joint id 7
            angle: 0 ~ 180
        """
        self.calibration_parameters(joint_id=joint_id, angle=angle)
        return self._mesg(ProtocolCode.SET_JOINT_MAX, joint_id, angle)

    def set_joint_min(self, joint_id, angle):
        """Set the joint minimum angle

        Args:
            joint_id: int.
                Joint id 1 - 6
                for gripper: Joint id 7
            angle: 0 ~ 180
        """
        self.calibration_parameters(joint_id=joint_id, angle=angle)
        return self._mesg(ProtocolCode.SET_JOINT_MIN, joint_id, angle)

    # servo control
    def is_servo_enable(self, servo_id):
        """To detect the connection state of a single joint

        Args:
            servo_id: 1 - 6
        Return:
            0 - disable
            1 - enable
            -1 - error
        """
        self.calibration_parameters(servo_id=servo_id)
        return self._mesg(ProtocolCode.IS_SERVO_ENABLE, servo_id, has_reply=True)

    def is_all_servo_enable(self):
        """Detect the connection status of all joints

        Return:
            0 - disable
            1 - enable
            -1 - error
        """
        return self._mesg(ProtocolCode.IS_ALL_SERVO_ENABLE, has_reply=True)

    def set_servo_data(self, servo_id, address, value, mode=None):
        """Set the data parameters of the specified address of the steering gear

        Args:
            servo_id: Serial number of articulated steering gear. 1 - 7
            address: Data address.
            value: 0 - 4096
            mode: 0 - indicates that value is one byte(default), 1 - 1 represents a value of two bytes.
        """

        unique_address = (0, 1, 2, 3, 4)
        if value in unique_address:
            raise ValueError(f"The address not right, Invalid address, address {unique_address} cannot be modified.")

        if mode is None:
            self.calibration_parameters(servo_id=servo_id, address=address, value=value)
            return self._mesg(ProtocolCode.SET_SERVO_DATA, servo_id, address, value)
        else:
            self.calibration_parameters(servo_id=servo_id, address=address, value=value, mode=mode)
            return self._mesg(ProtocolCode.SET_SERVO_DATA, servo_id, address, [value], mode)

    def get_servo_data(self, servo_id, address, mode=None):
        """Read the data parameter of the specified address of the steering gear.

        Args:
            servo_id: Serial number of articulated steering gear.1 - 7
            address: Data address.
            mode: 0 - indicates that value is one byte(default), 1 - 1 represents a value of two bytes.

        Return:
            values 0 - 4096
        """
        if mode is not None:
            self.calibration_parameters(servo_id=servo_id, address=address, mode=mode)
            return self._mesg(ProtocolCode.GET_SERVO_DATA, servo_id, address, mode, has_reply=True)
        self.calibration_parameters(servo_id=servo_id, address=address)
        return self._mesg(ProtocolCode.GET_SERVO_DATA, servo_id, address, has_reply=True)

    def set_servo_calibration(self, servo_id):
        """The current position of the calibration joint actuator is the angle zero point,
            and the corresponding potential value is 2048.

        Args:
            servo_id: Serial number of articulated steering gear. 1 - 7
        """
        self.calibration_parameters(servo_id=servo_id)
        return self._mesg(ProtocolCode.SET_SERVO_CALIBRATION, servo_id)

    def joint_brake(self, joint_id):
        """Make it stop when the joint is in motion, and the buffer distance is positively related to the existing speed

        Args:
            joint_id:  1 - 6
        """
        self.calibration_parameters(joint_id=joint_id)
        return self._mesg(ProtocolCode.JOINT_BRAKE, joint_id)

    def release_servo(self, servo_id):
        """Power off designated servo

        Args:
            servo_id: int 1 - 7
        """
        self.calibration_parameters(servo_id=servo_id)
        return self._mesg(ProtocolCode.RELEASE_SERVO, servo_id)

    def focus_servo(self, servo_id):
        """Power on designated servo

        Args:
            servo_id: int 1 - 7
        """
        self.calibration_parameters(servo_id=servo_id)
        return self._mesg(ProtocolCode.FOCUS_SERVO, servo_id)

    # 夹爪控制
    def get_gripper_value(self, gripper_type=1):
        """Get the value of gripper.

        Args:
            gripper_type (int): default 1
                1: Adaptive gripper
                3: Parallel gripper
                4: Flexible gripper

        Return:
            gripper value (int)
        """
        self.calibration_parameters(gripper_type=gripper_type)
        return self._mesg(ProtocolCode.GET_GRIPPER_VALUE, gripper_type, has_reply=True)

    def set_gripper_state(self, gripper_state, gripper_speed, gripper_type=None, is_torque=None):
        """Set gripper switch state

        Args:
            gripper_state  (int): 0 - open, 1 - close, 254 - release
            gripper_speed (int): 1 ~ 100
            gripper_type (int): default 1
                1 : Adaptive gripper. default to adaptive gripper
                2 : 5 finger dexterous hand
                3 : Parallel gripper, this parameter can be omitted
                4 : Flexible gripper
            is_torque (int): When there is no type parameter, this parameter can be omitted.
                1: Force control
                0: Non-force control
        """
        self.calibration_parameters(
            gripper_state=gripper_state, gripper_speed=gripper_speed, gripper_type=gripper_type, is_torque=is_torque
        )
        args = [gripper_state, gripper_speed]
        if gripper_type is not None:
            args.append(gripper_type)
        if is_torque is not None:
            args.append(is_torque)
        return self._mesg(ProtocolCode.SET_GRIPPER_STATE, *args)

    def set_gripper_value(self, gripper_value, gripper_speed, gripper_type=None, is_torque=None):
        """Set gripper value

        Args:
            gripper_value (int): 0 ~ 100
            gripper_speed (int): 1 ~ 100
            gripper_type (int): default 1
                1: Adaptive gripper
                3: Parallel gripper, this parameter can be omitted
                4: Flexible gripper
            is_torque (int): When there is no type parameter, this parameter can be omitted.
                1: Force control
                0: Non-force control
        """
        self.calibration_parameters(
            gripper_value=gripper_value,
            speed=gripper_speed,
            gripper_type=gripper_type,
            is_torque=is_torque
        )
        args = [gripper_value, gripper_speed]
        if gripper_type is not None:
            args.append(gripper_type)
        if is_torque is not None:
            args.append(is_torque)
        return self._mesg(ProtocolCode.SET_GRIPPER_VALUE, *args, has_reply=True)

    def set_gripper_calibration(self):
        """Set the current position to zero, set current position value is `2048`."""
        return self._mesg(ProtocolCode.SET_GRIPPER_CALIBRATION)

    def is_gripper_moving(self):
        """Judge whether the gripper is moving or not

        Returns:
            0 - not moving
            1 - is moving
            -1- error data
        """
        return self._mesg(ProtocolCode.IS_GRIPPER_MOVING, has_reply=True)

    # Atom IO
    def get_tool_system_version(self):
        """
        Read the terminal primary and minor version numbers
        """
        return self._mesg(ProtocolCode.GET_ATOM_VERSION, has_reply=True)

    def get_tool_modify_version(self):
        """
        Read the terminal modified version number
        """
        return self._mesg(ProtocolCode.OVER_LIMIT_RETURN_ZERO, has_reply=True)

    def is_tool_connected(self):
        """Check the end connection status"""
        return self._mesg(ProtocolCode.IS_CONTROLLER_CONNECTED, has_reply=True)

    def set_tool_color(self, r=0, g=0, b=0):
        """Set the light color on the top of the robot arm.

        Args:
            r (int): 0 ~ 255
            g (int): 0 ~ 255
            b (int): 0 ~ 255

        """
        self.calibration_parameters(color=[r, g, b])
        return self._mesg(ProtocolCode.SET_COLOR, r, g, b)

    def is_tool_button_click(self):
        """Check whether the button on the end is pressed"""
        return self._mesg(ProtocolCode.IS_CONTROLLER_BUTTON_PRESSED, has_reply=True)

    def set_digital_output(self, pin_no, pin_signal):
        """Set the terminal atom io status

        Args:
            pin_no     (int): 1 or 2
            pin_signal (int): 0 / 1
        """

        if pin_no not in (1, 2):
            raise ValueError("pin_no must be 1 or 2")

        self.calibration_parameters(class_name=self.__class__.__name__, pin_signal=pin_signal)
        return self._mesg(ProtocolCode.SET_DIGITAL_OUTPUT, pin_no, pin_signal)

    def get_digital_input(self, pin_no):
        """Get atom digital input value

        Args:
            pin_no (int): 1 or 2
        """
        if pin_no not in (1, 2):
            raise ValueError("pin_no must be 1 or 2")

        return self._mesg(ProtocolCode.GET_DIGITAL_INPUT, pin_no, has_reply=True)

    # Kinematic algorithm interface api
    def solve_inv_kinematics(self, target_coords, current_angles):
        """ Convert target coordinates to angles

        Args:
            target_coords: A float list of all coordinates.
            current_angles : A float list of all angle.

        Return:
            list: A float list of all angle.
        """
        self.calibration_parameters(coords=target_coords, angles=current_angles)
        angles = [self._angle2int(angle) for angle in current_angles]
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(target_coords[idx]))
        for angle in target_coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SOLVE_INV_KINEMATICS, coord_list, angles, has_reply=True)

    # Coordinate transformation
    def set_tool_reference(self, coords):
        """Set tool coordinate system

        Args:
            coords: a list of coords value(List[float])
                    [x(mm), y, z, rx(angle), ry, rz]
        """
        self.calibration_parameters(coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_TOOL_REFERENCE, coord_list)

    def get_tool_reference(self):
        """Get tool coordinate system """
        return self._mesg(ProtocolCode.GET_TOOL_REFERENCE, has_reply=True)

    def set_world_reference(self, coords):
        """Set the world coordinate system

        Args:
            coords: a list of coords value(List[float])
                    [x(mm), y, z, rx(angle), ry, rz]\n
        """
        self.calibration_parameters(coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_WORLD_REFERENCE, coord_list)

    def get_world_reference(self):
        """Get the world coordinate system"""
        return self._mesg(ProtocolCode.GET_WORLD_REFERENCE, has_reply=True)

    def set_reference_frame(self, reference_frame_type):
        """Set the base coordinate system

        Args:
            reference_frame_type: 0 - base 1 - tool.
        """
        self.calibration_parameters(reference_frame_type=reference_frame_type)
        return self._mesg(ProtocolCode.SET_REFERENCE_FRAME, reference_frame_type)

    def get_reference_frame(self):
        """Get the base coordinate system

        Return:
            0 - base 1 - tool.
        """
        return self._mesg(ProtocolCode.GET_REFERENCE_FRAME, has_reply=True)

    def set_movement_type(self, move_type):
        """Set movement type

        Args:
            move_type: 1 - movel, 0 - moveJ
        """
        self.calibration_parameters(move_type=move_type)
        return self._mesg(ProtocolCode.SET_MOVEMENT_TYPE, move_type)

    def get_movement_type(self):
        """Get movement type

        Return:
            1 - movel, 0 - moveJ
        """
        return self._mesg(ProtocolCode.GET_MOVEMENT_TYPE, has_reply=True)

    def set_end_type(self, end_type):
        """Set end coordinate system

        Args:
            end_type: int
                0 - flange, 1 - tool
        """
        self.calibration_parameters(end_type=end_type)
        return self._mesg(ProtocolCode.SET_END_TYPE, end_type)

    def get_end_type(self):
        """Get end coordinate system

        Return:
            0 - flange, 1 - tool
        """
        return self._mesg(ProtocolCode.GET_END_TYPE, has_reply=True)

    # 9G Servo backgammon
    def move_round(self):
        """Drive the 9g steering gear clockwise for one revolution
        """
        return self._mesg(ProtocolCode.move_round)

    def set_four_pieces_zero(self):
        """Set the zero position of the four-piece motor

        Returns:
            int: 0 or 1 (1 - success)
        """
        return self._mesg(ProtocolCode.SET_FOUR_PIECES_ZERO, has_reply=True)

    # Stdio api
    def get_angles_coords(self):
        """Get joint angles and coordinates"""
        return self._mesg(ProtocolCode.GET_ANGLES_COORDS, has_reply=True)

    # Servo state value
    def get_servo_speeds(self):
        """Get joint speed

        Return:
            A list unit step/s
        """
        return self._mesg(ProtocolCode.GET_SERVO_SPEED, has_reply=True)

    def get_servo_currents(self):
        """Get all joint current

        Return:
             A list unit mA
        """
        return self._mesg(ProtocolCode.GET_SERVO_CURRENTS, has_reply=True)

    def get_servo_voltages(self):
        """Get joint voltages

        Return:
            A list volts < 24 V
        """
        return self._mesg(ProtocolCode.GET_SERVO_VOLTAGES, has_reply=True)

    def get_servo_status(self):
        """Get joint status

        Return:
            [voltage, sensor, temperature, current, angle, overload], a value of 0 means no error, a value of 1 indicates an error
        """
        return self._mesg(ProtocolCode.GET_SERVO_STATUS, has_reply=True)

    def get_servo_temps(self):
        """Get joint temperature

        Return:
            A list unit ℃
        """
        return self._mesg(ProtocolCode.GET_SERVO_TEMPS, has_reply=True)

    # Drag the teach-in
    def start_drag_record(self):
        """Start track recording

        Return:
            Recording queue length
        """

        return self._mesg(ProtocolCode.DRAG_START_RECORD, has_reply=True)

    def end_drag_record(self):
        """End track recording

        Return:
             Recording queue length
        """

        return self._mesg(ProtocolCode.DRAG_END_RECORD, has_reply=True)

    def get_drag_record_data(self):
        """Get the recorded track

        Return:
            List of potential values (encoder values) and operating speeds of each joint
            eg: [J1_encoder, J1_run_speed,J2_encoder, J2_run_speed,J3_encoder, J3_run_speed,J4_encoder, J4_run_speed,J5_
            encoder, J5_run_speed,J6_encoder, J6_run_speed]
        """

        return self._mesg(ProtocolCode.DRAG_GET_RECORD_DATA, has_reply=True)

    def get_drag_record_len(self):
        """Get the total number of recorded points

        Return:
            Recording queue length
        """

        return self._mesg(ProtocolCode.DRAG_GET_RECORD_LEN, has_reply=True)

    def clear_drag_record_data(self):
        """Clear recording track

        Return:
            Recording queue length 0
        """

        return self._mesg(ProtocolCode.DRAG_CLEAR_RECORD_DATA, has_reply=True)


class MyCobot280RDKX5(MyCobot280RDKX5CommandGenerator):

    def __init__(self, port, baudrate=100_0000, timeout=0.1, debug=False, thread_lock=True):
        """
        Args:
            port     : port string
            baudrate : baud rate int, default 100_0000
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super().__init__(debug, thread_lock)
        self._serial_port = setup_serial_port(port=port, baudrate=baudrate, timeout=timeout)
        self._write = functools.partial(write, self)
        self._read = functools.partial(read, self)

    def close(self):
        self._serial_port.close()

    def open(self):
        self._serial_port.open()


class MyCobot280RDKX5Socket(MyCobot280RDKX5CommandGenerator):
    """MyCobot 280 RDK X5 Socket Control Class

    server file: https://github.com/elephantrobotics/pymycobot/demo/Server_280_RDK_X5.py
    """
    def __init__(self, ip, port=30002, timeout=0.1, debug=False, thread_lock=True):
        super().__init__(debug, thread_lock)
        self.sock = setup_socket_connect(ip, port, timeout)
        self._write = functools.partial(write, self, method="socket")
        self._read = functools.partial(read, self, method="socket")

    def set_gpio_mode(self, mode):
        """Set pin coding method
        Args:
            mode: (int) 0 - BCM, 1 - BOARD

        returns:
            (int) 1 - success, 255 - error
        """
        if mode not in (0, 1):
            raise ValueError("mode must be 0 or 1")
        return self._mesg(GPIOProtocolCode.SETUP_GPIO_MODE, mode)

    def setup_gpio_state(self, pin_no, mode, initial=1):
        """Set the pin as input or output
        Args:
            pin_no: (int) pin id
            mode: (int) 0 - input, 1 - output
            initial: (int) 0 - low, 1 - high
        returns:
            (int) 1 - success, 255 - error
        """
        if mode not in (0, 1):
            raise ValueError("mode must be 0 or 1")

        if initial not in (0, 1):
            raise ValueError("initial must be 0 or 1")

        return self._mesg(GPIOProtocolCode.SETUP_GPIO_STATE, pin_no, mode, initial)

    def set_gpio_output(self, pin_no, state):
        """Set the pin to high or low level
        Args:
            pin_no: (int) pin id.
            state: (int) 0 - low, 1 - high
        returns:
            (int) 1 - success, 255 - error
        """
        return self._mesg(GPIOProtocolCode.SET_GPIO_OUTPUT, pin_no, state)

    def get_gpio_input(self, pin_no):
        """Get pin level status.
        Args:
            pin_no: (int) pin id.
        Returns:
            (int) 0 - low, 1 - high, 255 - error
        """
        return self._mesg(GPIOProtocolCode.GET_GPIO_INPUT, pin_no, has_reply=True)

    def close(self):
        self.sock.close()


def main():
    rdx_x5 = MyCobot280RDKX5("/dev/ttyS1", debug=True)
    # print(rdx_x5.get_angles())
    # rdx_x5.send_angle(6, 168, 100)
    rdx_x5.send_angles([0, 0, 0, 0, 0, 0], 100)
    time.sleep(3)
    # print(rdx_x5.get_coords())
    rdx_x5.send_coords([169.7, -65.8, 200.5, 178.14, -0.41, -89.98], 10)
    # mc_sock = MyCobot280RDKX5Socket('192.168.1.246', port=30002, debug=True)
    # print(mc_sock.send_angle(1, 100, 50))
    # print(mc_sock.get_quick_move_message())
    # print(mc_sock.set_gpio_mode(0))
    # print(mc_sock.setup_gpio_state(5, 1, initial=1))
    # print(mc_sock.set_gpio_output(5, 0))
    # print(mc_sock.get_gpio_input(5))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import decimal
import locale
import re
import threading
import time
import enum

from .myagvapi import CommunicationProtocol, Utils, setup_serial_connect, setup_logging

LOCAL_LANGUAGE_CODE, _ = locale.getdefaultlocale()


class ProtocolCode(enum.Enum):
    # System & Product Information
    GET_MODIFY_VERSION = 0x01
    GET_SYSTEM_VERSION = 0x02
    GET_ROBOT_STATUS = 0x05
    POWER_ON = 0x10
    POWER_ON_ONLY = 0x19
    POWER_OFF = 0x11
    GET_POWER_STATE = 0x12

    # Motion control
    AGV_MOTION_CONTROL = 0x21
    AGV_STOP_MOVING = 0x22
    SET_AUTO_REPORT_STATE = 0x23
    GET_AUTO_REPORT_STATE = 0x24
    GET_AUTO_REPORT_MESSAGE = 0x25

    # aided
    SET_MOTOR_ENABLED = 0x30
    GET_MOTOR_STATUS = 0x31
    SET_COMMUNICATION_MODE = 0x32
    GET_COMMUNICATION_MODE = 0x33
    SET_LED_COLOR = 0x34

    GET_MOTOR_TEMPERATURE = 0x35
    GET_MOTOR_SPEEDS = 0x36
    GET_MOTOR_TORQUES = 0x37
    GET_MOTOR_ENABLE_STATUS = 0x38
    GET_MOTOR_LOSS_COUNT = 0x39

    SET_OUTPUT_IO = 0x40
    GET_INPUT_IO = 0x41

    SET_LED_MODE = 0x3A

    GET_WIFI_ACCOUNT = 0x50
    GET_WIFI_IP_PORT = 0x51
    GET_BLUETOOTH_UUID = 0x52
    GET_BLUETOOTH_ADDRESS = 0x53

    SET_HANDLE_CONTROL_STATE = 0x3D
    GET_HANDLE_CONTROL_STATE = 0x3F

    def equal(self, other):
        if isinstance(other, ProtocolCode):
            return self.value == other.value
        else:
            return self.value == other


PLAINTEXT_REPLY_PROTOCOL_CODE = (
    ProtocolCode.GET_WIFI_ACCOUNT,
    ProtocolCode.GET_WIFI_IP_PORT,
    ProtocolCode.GET_BLUETOOTH_UUID,
    ProtocolCode.GET_BLUETOOTH_ADDRESS
)


class MyAGVProCommandProtocolApi(CommunicationProtocol):
    genre_timeout_table = {
        ProtocolCode.POWER_ON: 2.1
    }
    language_prompt_tips = {
        "en_US": {
            ProtocolCode.POWER_ON: {
                1: "[Power-on Tips]: Power on success",
                2: "[Power-on Tips]: Emergency stop triggered",
                3: "[Power-on Tips]: The battery is too low",
                4: "[Power-on Tips]: CAN initialization is abnormal",
                5: "[Power-on Tips]: Motor initialization exception"
            }
        },
        "zh_CN": {
            ProtocolCode.POWER_ON: {
                0: "[上电提示]: 上电成功",
                2: "[上电提示]: 紧急触发",
                3: "[上电提示]: 电量过低",
                4: "[上电提示]: CAN初始化异常",
                5: "[上电提示]: 电机初始化异常"
            }
        }
    }

    def __init__(self, debug=False, save_serial_log=False):
        self._mutex = threading.Lock()
        self._save_serial_log = save_serial_log
        self._data_buffer = []
        self._serial_filename = "serial.log"
        self._communication_mode = 0
        self.log = setup_logging(name=f"{__name__}.{self.__class__.__name__}", debug=debug)

    def _save_buffer_data(self, data, to_local=False):
        if self._save_serial_log is False:
            return

        self._data_buffer.append(data)

        if to_local is True:
            self._save_file(self._serial_filename, " ".join(map(str, self._data_buffer)) + "\n")
            self._data_buffer.clear()

    def _prompt(self, genre, code):
        default_prompt_tips = self.language_prompt_tips.get("en_US")
        code_prompt_tips = self.language_prompt_tips.get(LOCAL_LANGUAGE_CODE, default_prompt_tips)
        prompt_tips = code_prompt_tips.get(genre, None)
        if prompt_tips is None:
            return

        tips = prompt_tips.get(code, None)
        if tips is None:
            return

        print(tips)

    @staticmethod
    def get_significant_bit(number):
        return -decimal.Decimal(str(number)).as_tuple().exponent

    @staticmethod
    def _save_file(filename, data):
        with open(filename, "a+") as f:
            f.write(data)

    def _match_protocol_data(self, genre, timeout=0.1):
        is_save_mac_addr = False
        for _ in range(5):
            if ProtocolCode.SET_COMMUNICATION_MODE.equal(genre) and self._communication_mode == 2:
                data = self._read_plaintext_data(timeout=timeout)
                if len(data) == 0:
                    continue

                if is_save_mac_addr is False:
                    mac_addr = self._parsing_data(ProtocolCode.GET_BLUETOOTH_ADDRESS, data.decode('utf-8'))
                    info = f"MyAGVPro Bluetooth MAC Address: {mac_addr}"
                    self._save_file("AGVPro_BLUETOOTH_MAC_ADDR", info)
                    is_save_mac_addr = True
                    print(info)

                reply_data = self._read_protocol_data(timeout=timeout)
                if len(reply_data) == 0:
                    continue

                break

            if genre in PLAINTEXT_REPLY_PROTOCOL_CODE:
                reply_data = self._read_plaintext_data(timeout=timeout)
            else:
                reply_data = self._read_protocol_data(timeout=timeout)

                if reply_data and not genre.equal(reply_data[3]):  # check genre
                    continue

            if len(reply_data) == 0:
                continue

            self.log.info(f" read: {' '.join(f'{x:02x}' for x in reply_data)}")
            break
        else:
            reply_data = None
        return reply_data

    def _read_plaintext_data(self, timeout=0.1):
        reply_data = self._read_command_buffer(
            start_flag_caller=lambda: b'AGVPro:',
            end_flag_caller=lambda cmds: b'\r\n',
            timeout=timeout
        )
        return reply_data

    def _read_protocol_data(self, timeout=0.1):
        reply_data = self._read_command_buffer(
            start_flag_caller=lambda: b'\xfe\xfe',
            end_flag_caller=lambda cmds: Utils.crc16_check_bytes(cmds[:-2]),
            timeout=timeout
        )
        return reply_data

    def _merge(self, genre, *args):
        with self._mutex:
            self.clear()
            timeout = self.genre_timeout_table.get(genre, 0.1)
            if not ProtocolCode.GET_AUTO_REPORT_MESSAGE.equal(genre):
                real_command = self._combination(genre, args)
                self.log.info(f"write: {' '.join(f'{x:02x}' for x in real_command)}")
                self.write(real_command)

            reply_data = self._match_protocol_data(genre, timeout)
            decode_respond = self._instruction_decoding(genre, reply_data)
            data = self._parsing_data(genre, decode_respond)
            self._prompt(genre, data)
            self._save_buffer_data(b'', to_local=True)
            return data

    @classmethod
    def _filter(cls, data, target):
        for index in range(len(data) - 1, -1, -1):
            if data[index] != target:
                return data[:index + 1]
        return data

    @classmethod
    def _instruction_decoding(cls, genre, reply_data):
        if not reply_data:
            return None

        respond_body = reply_data[4: -2]
        if genre in PLAINTEXT_REPLY_PROTOCOL_CODE:
            respond_body = reply_data.decode("utf-8")

        if genre in (
                ProtocolCode.GET_MOTOR_TEMPERATURE,
                ProtocolCode.GET_MOTOR_SPEEDS,
                ProtocolCode.GET_MOTOR_TORQUES,
                ProtocolCode.GET_MOTOR_STATUS,
                ProtocolCode.GET_MOTOR_LOSS_COUNT
        ):
            respond = []
            for i in range(0, len(respond_body), 2):
                data = Utils.decode_int16(respond_body[i:i + 2])
                respond.append(data)
            return respond
        return respond_body

    @classmethod
    def _parsing_data(cls, genre, reply_data):
        if not reply_data:
            return None

        if ProtocolCode.GET_SYSTEM_VERSION.equal(genre):
            return reply_data[0] / 10

        if ProtocolCode.GET_MOTOR_TEMPERATURE.equal(genre):
            return list(data / 10 for data in reply_data)

        if genre in (ProtocolCode.GET_MOTOR_SPEEDS, ProtocolCode.GET_MOTOR_TORQUES):
            return list(data / 100 for data in reply_data)

        if ProtocolCode.GET_MOTOR_LOSS_COUNT.equal(genre):
            return list(reply_data)

        if ProtocolCode.GET_WIFI_ACCOUNT.equal(genre):
            data = re.findall(r'AGVPro:WIFI:S:(.*);P:(.*);', reply_data)
            return data[0] if len(data) == 1 else None

        if ProtocolCode.GET_WIFI_IP_PORT.equal(genre):
            data = re.findall(r'AGVPro:WIFI:IP:(.*);PORT:(.*);', reply_data)
            if len(data) == 0:
                return None
            return data[0][0], int(data[0][1])

        if ProtocolCode.GET_BLUETOOTH_UUID.equal(genre):
            data = re.findall(r'AGVPro:BLE::Name:(.*);Service_UUID:(.*);CHAR_UUID:(.*);', reply_data)
            return data[0] if len(data) == 1 else None

        if ProtocolCode.GET_BLUETOOTH_ADDRESS.equal(genre):
            data = re.findall(r'AGVPro:BLE:MAC:(.*);\r\n', reply_data)
            return data[0] if len(data) == 1 else None

        if ProtocolCode.GET_MOTOR_STATUS.equal(genre):
            respond = []
            for item in reply_data:
                if item == 0:
                    respond.append(0)
                    continue

                rank = Utils.get_bits(item)
                respond.append(rank)
            return respond

        if ProtocolCode.GET_ROBOT_STATUS.equal(genre):
            machine_states = [0] * 8
            machine_status = reply_data[0]
            if machine_status != 0:
                for index in Utils.get_bits(machine_status):
                    machine_states[index] = 1

            battery_voltage = round(reply_data[1] / 10, 2)

            return machine_states, battery_voltage

        if ProtocolCode.GET_AUTO_REPORT_MESSAGE.equal(genre):
            respond = []
            for index, data in enumerate(reply_data):
                if index < 3:
                    respond.append(round(data / 100, 2))

                elif index in (3, 4):
                    if data == 0:
                        rank = data
                    else:
                        rank = Utils.get_bits(data)
                    respond.append(rank)

                elif index == 5:
                    respond.append(round(data / 10, 1))

                elif index == 6:
                    respond.append(data)

                elif index > 7 and index % 2 == 0:
                    piece = reply_data[index - 1:index + 1]
                    int16 = Utils.decode_int16(piece)
                    respond.append(round(int16 / 100, 2))

            return respond

        if ProtocolCode.GET_MOTOR_ENABLE_STATUS.equal(genre):
            return list(reply_data[:4])

        if ProtocolCode.GET_INPUT_IO.equal(genre):
            if reply_data[0] == 255:
                return -1
            return reply_data[1]

        return reply_data[0]

    @classmethod
    def _combination(cls, genre, params):
        command_args = Utils.process_data_command(params)
        while len(command_args) < 8:
            command_args.append(0x00)
        command = bytearray([0xfe, 0xfe, 0x0b, genre.value])
        command.extend(command_args)
        crc16_check_code = Utils.crc16_check(command)

        command.extend(crc16_check_code)
        return command

    def _read_by_timeout(self, timeout=1.0):
        start_time = time.time()
        while time.time() - start_time < timeout:
            data = self.read(1)
            if len(data) == 0:
                continue

            yield data

            self._save_buffer_data(data)

    def _read_command_buffer(self, start_flag_caller, end_flag_caller, timeout=1.0):
        channel_buffer = b""
        real_command = b""
        is_record = False

        for data in self._read_by_timeout(timeout):
            channel_buffer += data

            start_flag = start_flag_caller()
            if channel_buffer.endswith(start_flag) and is_record is False:
                real_command = start_flag
                is_record = True
                continue

            if is_record is False:
                continue

            real_command += data
            end_flag = end_flag_caller(real_command)
            if real_command.endswith(end_flag):
                break
        else:
            real_command = b""
        return real_command


class MyAGVProCommandApi(MyAGVProCommandProtocolApi):

    def get_system_version(self):
        """Obtain the major firmware version number

        Returns:
            float: version
        """
        return self._merge(ProtocolCode.GET_SYSTEM_VERSION)

    def get_modify_version(self):
        """Obtain the minor firmware version number

        Returns:
            int: version
        """
        return self._merge(ProtocolCode.GET_MODIFY_VERSION)

    def get_robot_status(self):
        """Obtain the machine status, and support the acquisition in the case of power failure

        Returns:
            tuple: (list[int] | int, float)
                0 - list[int] | int: machine status，Normally 0
                    0 - (int)Emergency stop status
                    1 - (int)Power status
                    2 - (int)Front bumper strip
                    3 - (int)Rear bumper strip
                    4 - (int)Motor No. 1 connection status
                    5 - (int)Motor No. 2 connection status
                    6 - (int)Motor No. 3 connection status
                    7 - (int)Motor No. 4 connection status
        """
        return self._merge(ProtocolCode.GET_ROBOT_STATUS)

    def power_on(self):
        """Turn on the robot

        Returns:
            int: Power-on result
                1 - Success
                2 - Emergency stop triggered
                3 - The battery is too low
                4 - CAN initialization is abnormal
                5 - Motor initialization exception
        """
        return self._merge(ProtocolCode.POWER_ON)

    def power_on_only(self):
        """Turn on the robot, but not start the control program

        Returns:
            int: Power-on result only, 1: Success, 0: Failed
        """
        return self._merge(ProtocolCode.POWER_ON_ONLY)

    def power_off(self):
        """Turn off the robot

        Returns:
            int: Power-off result only, 1: Success, 0: Failed
        """
        return self._merge(ProtocolCode.POWER_OFF)

    def is_power_on(self):
        """Check if the robot is powered on

        Returns:
            int: power state, 1: Power-on, 0: Power-off
        """
        return self._merge(ProtocolCode.GET_POWER_STATE)

    def _basic_move(self, vertical_speed=0.0, horizontal_speed=0.0, rotate_speed=0.0):
        """ Basic moving control

        Args:
            vertical_speed: +-0.01 ~ +-1.50m/s
            horizontal_speed: +-0.01 ~ +-1.00m/s
            rotate_speed:

        Returns:
            int: 1: Success, 0: Failed
        """
        return self._merge(
            ProtocolCode.AGV_MOTION_CONTROL,
            [
                int(vertical_speed * 100),
                int(horizontal_speed * 100),
                int(rotate_speed * 100)
            ]
        )

    def move_forward(self, speed):
        """Pan the robot forward

        Args:
            speed(float): 0.01 ~ 1.50m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.50:
            raise ValueError("Speed must be between 0.01 and 1.50")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return self._merge(ProtocolCode.AGV_MOTION_CONTROL, [int(speed * 100 * 1), 0x00])

    def move_backward(self, speed):
        """Pan the robot backward

        Args:
            speed(float): 0.01 ~ 1.50m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.50:
            raise ValueError("Speed must be between 0.01 and 1.50")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")

        return self._merge(ProtocolCode.AGV_MOTION_CONTROL, [int(speed * 100 * -1)])

    def move_left_lateral(self, speed):
        """Pan the robot left

        Args:
            speed(float): 0.01 ~ 1.00 m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.00:
            raise ValueError("Speed must be between 0.01 and 1.00")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, int(speed * 100 * -1)])

    def move_right_lateral(self, speed):
        """Pan the robot right

        Args:
            speed(float): 0.01 ~ 1.00m/s

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 0.01 <= speed <= 1.00:
            raise ValueError("Speed must be between 0.01 and 1.00")

        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, int(speed * 100 * 1)])

    def turn_left(self, speed):
        """Rotate to the left

        Args:
            speed:

        Returns:
            int: 1: Success, 0: Failed
        """
        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, 0x00, int(speed * 100 * -1), 0x00])

    def turn_right(self, speed):
        """Rotate to the right

        Args:
            speed:

        Returns:
            int: 1: Success, 0: Failed
        """
        if self.get_significant_bit(speed) > 2:
            raise ValueError(f"speed must be a number with 2 significant bits, but got {speed}")
        return self._merge(ProtocolCode.AGV_MOTION_CONTROL, [0x00, 0x00, int(speed * 100 * 1)])

    def stop(self):
        """Stop moving

        Returns:
            int: 1: Success, 0: Failed
        """
        return self._merge(ProtocolCode.AGV_STOP_MOVING)

    def set_auto_report_state(self, state):
        """Set the auto-report state

        Args:
            state(int): 0: Close, 1: Open

        Returns:
            int: 1: Success, 0: Failed
        """

        if state not in (0, 1):
            raise ValueError("State must be 0 or 1")

        return self._merge(ProtocolCode.SET_AUTO_REPORT_STATE, state)

    def get_auto_report_state(self):
        """Get the auto-report state

        Returns:
            int: 0: Close, 1: Open
        """
        return self._merge(ProtocolCode.GET_AUTO_REPORT_STATE)

    def get_auto_report_message(self):
        """Get the auto-report message
        Returns:
            list[int | list[int] | float]:
             0 - (float)rx
             1 - (float)ry
             2 - (float)rw
             3 - (list[int] | int)Machine status, 0 in normal cases and a list in abnormal cases
                0 - (int)Emergency stop status
                1 - (int)Power status
                2 - (int)Front bumper strip
                3 - (int)Rear bumper strip
             4 - (list[int] | int)Motor information, It is 0 when it is normal, and returns the motor number list when it is abnormal
             5 - (float)Battery voltage
             6 - (int)Motor enable status 0: Enabled, 1: Disabled
        """
        return self._merge(ProtocolCode.GET_AUTO_REPORT_MESSAGE)

    def set_motor_enable(self, motor_id, state):
        """Enable or disable the motor

        Args:
            motor_id(int):
                1: Left upper motor
                2: Right upper motor
                3: Left lower motor
                4: Right lower motor
                254: All motors
            state(bool):
                0: Disable
                1: Enable
        Returns:
            int: 1: Success, 0: Failed
        """
        if motor_id not in (1, 2, 3, 4, 254):
            raise ValueError("Motor id must be in (1, 2, 3, 4, 254)")

        if state not in (0, 1):
            raise ValueError("State must be 0 or 1")

        return self._merge(ProtocolCode.SET_MOTOR_ENABLED, motor_id, state)

    def get_motor_status(self):
        """Get the motor status

        Returns:
            list[int]: Motor status
                0: normal
                any: error code
        """
        return self._merge(ProtocolCode.GET_MOTOR_STATUS)

    def get_motor_temps(self):
        """Get the motor temperature

        Returns:
            list[float]: Motor temperature
        """
        return self._merge(ProtocolCode.GET_MOTOR_TEMPERATURE)

    def get_motor_speeds(self):
        """Get the motor speeds

        Returns:
            list[float]: Motor speeds
        """
        return self._merge(ProtocolCode.GET_MOTOR_SPEEDS)

    def get_motor_torques(self):
        """Get the motor torques

        Returns:
            list[float]: Motor torques
        """
        return self._merge(ProtocolCode.GET_MOTOR_TORQUES)

    def get_motor_enable_status(self):
        """Get the motor enabled status
        Returns:
            list[int]: Motor enabled status
                0: Disable
                1: Enable
        """
        return self._merge(ProtocolCode.GET_MOTOR_ENABLE_STATUS)

    def get_motor_loss_count(self):
        """Get the motor loss count

        Returns:
            list[int]: Motor loss count
        """
        return self._merge(ProtocolCode.GET_MOTOR_LOSS_COUNT)

    def set_communication_state(self, state):
        """Set the communication state

        Args:
            state(int):
                0: Serial communication (default)
                1: Socket communication
                2: Bluetooth communication (Write the MAC address to the file and the endpoint, and then return to the state)

        Returns:
            int: 1: Success, 0: Failed
        """
        if state not in (0, 1, 2):
            raise ValueError("State must be 0, 1 or 2")

        self._communication_mode = state
        return self._merge(ProtocolCode.SET_COMMUNICATION_MODE, state)

    def get_communication_state(self):
        """Get the communication state

        Returns:
            int: communication state
                0: Serial communication,
                1: Socket communication,
                2: Bluetooth communication
        """
        return self._merge(ProtocolCode.GET_COMMUNICATION_MODE)

    def set_led_color(self, position, color, brightness=255):
        """Set the LED color

        Args:
            position(int):
                0: Left LED
                1: Right LED
            color(tuple(int, int, int)): RGB color
            brightness(int): 0 - 255 (default: 255)

        Returns:
            int: 1: Success, 0: Failed
        """
        if position not in (0, 1):
            raise ValueError("Position must be 0 or 1")

        if not 0 <= brightness <= 255:
            raise ValueError("Brightness must be between 0 and 255")

        if len(color) != 3:
            raise ValueError("Color must be a tuple of 3 values")

        if any(map(lambda c: not 0 <= c <= 255, color)):
            raise ValueError("Color must be between 0 and 255")

        return self._merge(ProtocolCode.SET_LED_COLOR, position, brightness, *color)

    def set_led_mode(self, mode):
        """Set the LED mode

        Args:
            mode(int):
                0: Battery level(default)
                1: DIY

        Returns:
            int: 1: Success, 0: Failed
        """
        if mode not in (0, 1):
            raise ValueError("Mode must be 0 or 1")
        return self._merge(ProtocolCode.SET_LED_MODE, mode)

    def set_pin_output(self, pin, state):
        """Set the output IO

        Args:
            pin(int): 1 - 6
            state(int): 0: Low, 1: High

        Returns:
            int: 1: Success, 0: Failed
        """
        if not 1 <= pin <= 6:
            raise ValueError("Pin must be between 1 and 6")

        if state not in (0, 1):
            raise ValueError("State must be 0 or 1")

        return self._merge(ProtocolCode.SET_OUTPUT_IO, pin, state)

    def get_pin_input(self, pin):
        """Get the input IO

        Args:
            pin(int): 1 - 6

        Returns:
            int: 0: Low, 1: High, -1: There is no such pin
        """
        if not 1 <= pin <= 6:
            raise ValueError("Pin must be between 1 and 6")
        return self._merge(ProtocolCode.GET_INPUT_IO, pin)

    def get_estop_state(self):
        """Get the emergency stop state

        Returns:
            int: 0: Release, 1: Press
        """
        return self._merge(ProtocolCode.GET_INPUT_IO, 254)

    def get_wifi_account(self):
        """Get the wi-fi account

        Returns:
            tuple(str, str): wi-fi account, wi-fi password
        """
        return self._merge(ProtocolCode.GET_WIFI_ACCOUNT)

    def get_wifi_ip_port(self):
        """Get the wi-fi ip and port
        Returns:
            tuple(str, int): wi-fi ip, wi-fi port
        """
        return self._merge(ProtocolCode.GET_WIFI_IP_PORT)

    def get_bluetooth_uuid(self):
        """Get the bluetooth uuid

        Returns:
            tuple(str, str, str): bluetooth name, service uuid, characteristic uuid
        """

        return self._merge(ProtocolCode.GET_BLUETOOTH_UUID)

    def get_bluetooth_address(self):
        """Get the bluetooth MAC address

        Returns:
            str: bluetooth MAC address
        """

        return self._merge(ProtocolCode.GET_BLUETOOTH_ADDRESS)

    def set_handle_control_state(self, state):
        """Set the handle control switch status
        Args:
            state(int): 0: Disable, 1: Enable

        Returns:
            int: 1: Success, 0: Failed
        """
        if state not in (0, 1):
            raise ValueError("state must be 0 or 1")
        return self._merge(ProtocolCode.SET_HANDLE_CONTROL_STATE, state)

    def get_handle_control_state(self):
        """Get the handle control switch status

        Returns:
            int: 0: Disable, 1: Enable
        """
        return self._merge(ProtocolCode.GET_HANDLE_CONTROL_STATE)


class MyAGVPro(MyAGVProCommandApi):

    def __init__(self, port, baudrate=1000000, timeout=0.1, debug=False, save_serial_log=False):
        super().__init__(debug=debug, save_serial_log=save_serial_log)
        self._serial = setup_serial_connect(port=port, baudrate=baudrate, timeout=timeout)
        self._serial_filename = 'agvpro_serial_serial.log'

    def write(self, command):
        return self._serial.write(command)

    def read(self, size=1):
        return self._serial.read(size)

    def close(self):
        if self._serial.is_open:
            self._serial.close()

    def open(self):
        if not self._serial.is_open:
            self._serial.open()

    def is_open(self):
        return self._serial.is_open

    def clear(self):
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()



from enum import Enum


class Axis(Enum):
    X = 0
    Y = 1
    Z = 2
    RX = 3
    RY = 4
    RZ = 5


class Joint(Enum):
    J1 = 0
    J2 = 1
    J3 = 2
    J4 = 3
    J5 = 4
    J6 = 5


class DI(Enum):
    """Available Digital Input Pins."""

    PIN_0 = 0
    PIN_1 = 1
    PIN_2 = 2
    PIN_3 = 3
    PIN_4 = 4
    PIN_5 = 5
    PIN_6 = 6
    PIN_7 = 7
    PIN_8 = 8
    PIN_9 = 9
    PIN_10 = 10
    PIN_11 = 11
    PIN_12 = 12
    PIN_13 = 13
    PIN_14 = 14

    CAT_PHYSICAL_START = 15
    CAT_PHYSICAL_STOP = 16
    CAT_PHYSICAL_USER_DEFINE = 17

    PIN_18 = 18
    PIN_19 = 19
    PIN_20 = 20
    PIN_21 = 21
    PIN_22 = 22
    PIN_23 = 23
    PIN_24 = 24
    PIN_25 = 25
    PIN_26 = 26
    PIN_27 = 27
    PIN_28 = 28
    PIN_29 = 29
    PIN_30 = 30
    PIN_31 = 31
    PIN_32 = 32
    PIN_33 = 33
    COLLISION_DETECTED = 33

    J1_COMMUNICATION = 34
    J2_COMMUNICATION = 35
    J3_COMMUNICATION = 36
    J4_COMMUNICATION = 37
    J5_COMMUNICATION = 38
    J6_COMMUNICATION = 39

    IO_STOP_TRIGGERED = 40
    IO_RUN_TRIGGERED = 41

    PIN_42 = 42

    HARDWARE_PAUSE_PRESSED = 43

    BRAKE_ACTIVATION_RUNNING = 44
    J1_SERVO_ENABLED = 45
    J2_SERVO_ENABLED = 46
    J3_SERVO_ENABLED = 47
    J4_SERVO_ENABLED = 48
    J5_SERVO_ENABLED = 49
    J6_SERVO_ENABLED = 50

    HARDWARE_FREE_MOVE = 51
    MOTION_ENABLE = 52

    PIN_53 = 53
    PIN_54 = 54
    PIN_55 = 55

    J1_STATUS = 56
    J2_STATUS = 57
    J3_STATUS = 58
    J4_STATUS = 59
    J5_STATUS = 60
    J6_STATUS = 61

    EMERGENCY_STOP = 62
    POWER_ON_STATUS = 63


class DO(Enum):
    """Available Digital Output Pins."""

    PIN_0 = 0
    PIN_1 = 1
    PIN_2 = 2
    PIN_3 = 3
    PIN_4 = 4
    PIN_5 = 5
    PIN_6 = 6
    PIN_7 = 7
    PIN_8 = 8
    PIN_9 = 9
    PIN_10 = 10
    PIN_11 = 11
    PIN_12 = 12
    PIN_13 = 13
    PIN_14 = 14
    PIN_15 = 15
    PIN_16 = 16
    PIN_17 = 17
    PIN_18 = 18
    PIN_19 = 19
    PIN_20 = 20
    PIN_21 = 21
    PIN_22 = 22
    PIN_23 = 23
    PIN_24 = 24
    PIN_25 = 25
    PIN_26 = 26
    PIN_27 = 27
    PIN_28 = 28
    PIN_29 = 29
    PIN_30 = 30
    PIN_31 = 31
    PIN_32 = 32
    PIN_33 = 33
    PIN_34 = 34
    PIN_35 = 35
    PIN_36 = 36
    PIN_37 = 37
    PIN_38 = 38
    PIN_39 = 39
    PIN_40 = 40
    PIN_41 = 41
    PIN_42 = 42
    PIN_43 = 43
    PIN_44 = 44
    PIN_45 = 45
    PIN_46 = 46

    BRAKE_ACTIVE_AUTO = 47
    BRAKE_MANUAL_MODE_ENABLE = 48
    J1_BRAKE_RELEASE = 49
    J2_BRAKE_RELEASE = 50
    J3_BRAKE_RELEASE = 51
    J4_BRAKE_RELEASE = 52
    J5_BRAKE_RELEASE = 53
    J6_BRAKE_RELEASE = 54

    SKIP_INIT_ERRORS = 55
    PROGRAM_AUTO_RUNNING = 56
    POWER_ON_RELAY_1 = 57
    POWER_ON_RELAY_2 = 58

    PIN_59 = 59

    SOFTWARE_FREE_MOVE = 60

    PIN_61 = 61
    PIN_62 = 62
    PIN_63 = 63


class AI(Enum):
    """Available Analog Input Pins."""

    PIN_0 = 0
    PIN_1 = 1
    PIN_2 = 2
    PIN_3 = 3
    PIN_4 = 4
    PIN_5 = 5
    PIN_6 = 6
    PIN_7 = 7
    PIN_8 = 8
    PIN_9 = 9
    PIN_10 = 10
    PIN_11 = 11
    PIN_12 = 12
    PIN_13 = 13
    PIN_14 = 14
    PIN_15 = 15
    PIN_16 = 16
    PIN_17 = 17
    PIN_18 = 18
    PIN_19 = 19
    PIN_20 = 20
    PIN_21 = 21
    PIN_22 = 22
    PIN_23 = 23
    PIN_24 = 24

    PIN_25 = 25
    AXIS_X_VELOCITY = 25
    PIN_26 = 26
    AXIS_Y_VELOCITY = 26
    PIN_27 = 27
    AXIS_Z_VELOCITY = 27
    PIN_28 = 28
    AXIS_RX_VELOCITY = 28
    PIN_29 = 29
    AXIS_RY_VELOCITY = 29
    PIN_30 = 30
    AXIS_RZ_VELOCITY = 30

    J1_VOLTAGE = 31
    J2_VOLTAGE = 32
    J3_VOLTAGE = 33
    J4_VOLTAGE = 34
    J5_VOLTAGE = 35
    J6_VOLTAGE = 36

    J1_TEMPERATURE = 37
    J2_TEMPERATURE = 38
    J3_TEMPERATURE = 39
    J4_TEMPERATURE = 40
    J5_TEMPERATURE = 41
    J6_TEMPERATURE = 42

    J1_WINDING_A_CURRENT = 43
    J2_WINDING_A_CURRENT = 44
    J3_WINDING_A_CURRENT = 45
    J4_WINDING_A_CURRENT = 46
    J5_WINDING_A_CURRENT = 47
    J6_WINDING_A_CURRENT = 48

    J1_WINDING_B_CURRENT = 49
    J2_WINDING_B_CURRENT = 50
    J3_WINDING_B_CURRENT = 51
    J4_WINDING_B_CURRENT = 52
    J5_WINDING_B_CURRENT = 53
    J6_WINDING_B_CURRENT = 54

    ROBOT = 55

    ROBOT_AVG_POWER = 56
    CONTROLLER_TEMPERATURE = 57

    J1_CURRENT = 58
    J2_CURRENT = 59
    J3_CURRENT = 60
    J4_CURRENT = 61
    J5_CURRENT = 62
    J6_CURRENT = 63


class AO(Enum):
    """Available Analog Output Pins."""

    PIN_0 = 0
    PIN_1 = 1
    PIN_2 = 2
    PIN_3 = 3
    PIN_4 = 4
    PIN_5 = 5
    PIN_6 = 6
    PIN_7 = 7
    PIN_8 = 8
    PIN_9 = 9
    PIN_10 = 10
    PIN_11 = 11
    PIN_12 = 12
    PIN_13 = 13
    PIN_14 = 14
    PIN_15 = 15

    LED_LIGHT = 16

    PIN_17 = 17
    PIN_18 = 18
    PIN_19 = 19
    PIN_20 = 20
    PIN_21 = 21
    PIN_22 = 22
    PIN_23 = 23
    PIN_24 = 24
    PIN_25 = 25
    PIN_26 = 26
    PIN_27 = 27
    PIN_28 = 28
    PIN_29 = 29
    PIN_30 = 30
    PIN_31 = 31
    PIN_32 = 32
    PIN_33 = 33
    PIN_34 = 34
    PIN_35 = 35
    PIN_36 = 36
    PIN_37 = 37
    PIN_38 = 38
    PIN_39 = 39
    PIN_40 = 40

    ACCELERATION = 41

    PIN_42 = 42
    PIN_43 = 43
    PIN_44 = 44
    PIN_45 = 45
    PIN_46 = 46
    PIN_47 = 47

    TOOL_FORCE = 48
    TOOL_SPEED = 49
    STOPPING_DISTANCE = 50
    STOPPING_TIME = 51
    POWER_LIMIT = 52

    PAYLOAD = 53

    PIN_54 = 54
    PIN_55 = 55

    J1_TORQUE = 56
    J2_TORQUE = 57
    J3_TORQUE = 58
    J4_TORQUE = 59
    J5_TORQUE = 60
    J6_TORQUE = 61

    XY_AXIS_TORQUE = 62
    Z_AXIS_TORQUE = 63



# coding=utf-8
import locale
import time
import threading
import serial
from pymycobot.close_loop import CloseLoop
from pymycobot.error import calibration_parameters
from pymycobot.common import ProtocolCode
from pymycobot.robot_info import _interpret_status_code


def get_local_language():
    language, _ = locale.getdefaultlocale()
    if language not in ["zh_CN", "en_US"]:
        language = "en_US"
    return language


def setup_serial_connect(port, baudrate="115200", timeout=0.1):
    serial_api = serial.Serial()
    serial_api.port = port
    serial_api.baudrate = baudrate
    serial_api.timeout = timeout
    serial_api.rts = False
    serial_api.open()
    return serial_api


class Pro630Api(CloseLoop):
    joint_number = 6
    power_control_1 = 3
    power_control_2 = 4
    arm_span_maximum = 630
    pin_numbered_mapping_table = {
        1: 17,
        2: 27,
        3: 22,
        4: 5,
        5: 6,
        6: 19
    }

    def __init__(self, debug=False, save_serial_log=False, method=None):
        super(Pro630Api, self).__init__(debug)
        self._joint_max_angles = [0] * self.joint_number
        self._joint_min_angles = [0] * self.joint_number
        self.save_serial_log = save_serial_log
        try:
            import numpy as np
        except ImportError:
            raise ImportError("Please install numpy")

        self.calibration_parameters = calibration_parameters
        self.language = get_local_language()
        self.read_threading = threading.Thread(target=self.read_thread, args=(method, ))
        self.read_threading.daemon = True
        self.read_threading.start()

    def _decode_command(self, genre, read_data):
        if read_data is None:
            return None

        if isinstance(read_data, int):
            return read_data

        decode_respond = []

        valid_data, data_len = read_data

        if data_len in [8, 12, 14, 16, 26, 60]:
            if data_len == 8 and (genre == ProtocolCode.IS_INIT_CALIBRATION):
                if valid_data[0] == 1:
                    return 1
                n = len(valid_data)
                for v in range(1, n):
                    decode_respond.append(valid_data[v])
            elif data_len == 8 and genre == ProtocolCode.GET_DOWN_ENCODERS:
                decode_respond = self.bytes4_to_int(valid_data)
            elif data_len == 6 and genre in [ProtocolCode.GET_SERVO_STATUS, ProtocolCode.GET_SERVO_VOLTAGES,
                                             ProtocolCode.GET_SERVO_CURRENTS]:
                for i in range(data_len):
                    decode_respond.append(valid_data[i])
            elif genre == ProtocolCode.MERCURY_ROBOT_STATUS:
                i = 0
                res = []
                while i < data_len:
                    if i < 8:
                        res.append(valid_data[i])
                        i += 1
                    else:
                        one = valid_data[i: i + 2]
                        res.append(self._decode_int16(one))
                        i += 2
                return res
            else:
                for header_i in range(0, len(valid_data), 2):
                    one = valid_data[header_i: header_i + 2]
                    decode_respond.append(self._decode_int16(one))
        elif data_len == 2:
            if genre in [ProtocolCode.IS_SERVO_ENABLE]:
                return [self._decode_int8(valid_data[1:2])]
            elif genre in [ProtocolCode.GET_ERROR_INFO]:
                return [self._decode_int8(valid_data[1:])]
            elif genre in [ProtocolCode.GET_LIMIT_SWITCH]:
                for i in valid_data:
                    decode_respond.append(i)
            elif genre in [ProtocolCode.GET_ROBOT_ID]:
                decode_respond.append(self._decode_int16(valid_data))
        elif data_len == 3:
            decode_respond.append(self._decode_int16(valid_data[1:]))
        elif data_len == 4:
            if genre == ProtocolCode.COBOTX_GET_ANGLE:
                decode_respond = self.bytes4_to_int(valid_data)
            for i in range(1, 4):
                decode_respond.append(valid_data[i])
        elif data_len == 7:
            error_list = [i for i in valid_data]
            for i in error_list:
                if i in range(16, 23):
                    decode_respond.append(1)
                elif i in range(23, 29):
                    decode_respond.append(2)
                elif i in range(32, 112):
                    decode_respond.append(3)
                else:
                    decode_respond.append(i)
        elif data_len == 24:
            decode_respond = self.bytes4_to_int(valid_data)
        elif data_len == 40:
            i = 0
            while i < data_len:
                if i < 28:
                    decode_respond += self.bytes4_to_int(valid_data)
                    i += 4
                else:
                    one = valid_data[i: i + 2]
                    decode_respond.append(self._decode_int16(one))
                    i += 2
        elif data_len == 32:
            i = 0
            res = []
            while i < 30:
                if i < 9 or i >= 23:
                    res.append(valid_data[i])
                    i += 1
                elif i < 23:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
            return res
        elif data_len == 38:
            i = 0
            res = []
            while i < data_len:
                if i < 10 or i >= 30:
                    res.append(valid_data[i])
                    i += 1
                elif i < 38:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
            return res
        elif data_len == 56:
            for i in range(0, data_len, 8):
                byte_value = int.from_bytes(valid_data[i:i + 4], byteorder='big', signed=True)
                decode_respond.append(byte_value)
        elif data_len == 6:
            for i in valid_data:
                decode_respond.append(i)
        else:
            if genre in [
                ProtocolCode.GET_SERVO_VOLTAGES,
                ProtocolCode.GET_SERVO_STATUS,
                ProtocolCode.GET_SERVO_TEMPS,
            ]:
                for i in range(data_len):
                    data1 = self._decode_int8(valid_data[i: i + 1])
                    decode_respond.append(0xFF & data1 if data1 < 0 else data1)
            decode_respond.append(self._decode_int8(valid_data))

        return decode_respond

    def _parsing_parameters(self, genre, res):
        if res is None:
            return None

        if isinstance(res, int):
            return res

        if genre in [
            ProtocolCode.ROBOT_VERSION,
            ProtocolCode.GET_ROBOT_ID,
            ProtocolCode.IS_POWER_ON,
            ProtocolCode.IS_CONTROLLER_CONNECTED,
            ProtocolCode.IS_PAUSED,
            ProtocolCode.IS_IN_POSITION,
            ProtocolCode.IS_MOVING,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.IS_ALL_SERVO_ENABLE,
            ProtocolCode.GET_SERVO_DATA,
            ProtocolCode.GET_DIGITAL_INPUT,
            ProtocolCode.GET_GRIPPER_VALUE,
            ProtocolCode.IS_GRIPPER_MOVING,
            ProtocolCode.GET_SPEED,
            ProtocolCode.GET_ENCODER,
            ProtocolCode.GET_BASIC_INPUT,
            ProtocolCode.GET_TOF_DISTANCE,
            ProtocolCode.GET_END_TYPE,
            ProtocolCode.GET_MOVEMENT_TYPE,
            ProtocolCode.GET_REFERENCE_FRAME,
            ProtocolCode.GET_FRESH_MODE,
            ProtocolCode.GET_GRIPPER_MODE,
            ProtocolCode.SET_SSID_PWD,
            ProtocolCode.GET_ERROR_DETECT_MODE,
            ProtocolCode.POWER_ON,
            ProtocolCode.POWER_OFF,
            ProtocolCode.RELEASE_ALL_SERVOS,
            ProtocolCode.RELEASE_SERVO,
            ProtocolCode.FOCUS_ALL_SERVOS,
            ProtocolCode.FOCUS_SERVO,
            ProtocolCode.STOP,
            ProtocolCode.SET_BREAK,
            ProtocolCode.IS_BTN_CLICKED,
            ProtocolCode.GET_CONTROL_MODE,
            ProtocolCode.GET_VR_MODE,
            ProtocolCode.GET_FILTER_LEN,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.GET_POS_SWITCH,
            ProtocolCode.GET_ERROR_INFO
        ]:
            return self._process_single(res)
        elif genre in [ProtocolCode.GET_ANGLES, ProtocolCode.SOLVE_INV_KINEMATICS]:
            return [self._int3angle(angle) for angle in res]
        elif genre in [
            ProtocolCode.GET_COORDS,
            ProtocolCode.MERCURY_GET_BASE_COORDS,
            ProtocolCode.GET_TOOL_REFERENCE,
            ProtocolCode.GET_WORLD_REFERENCE,
        ]:
            if res:
                r = []
                for idx in range(3):
                    r.append(self._int2coord(res[idx]))
                for idx in range(3, 6):
                    r.append(self._int2angle(res[idx]))
                return r
            else:
                return res
        elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
            return [self._int2coord(angle) for angle in res]
        elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION]:
            return self._int2coord(self._process_single(res))
        elif genre in [
            ProtocolCode.GET_JOINT_MAX_ANGLE,
            ProtocolCode.GET_JOINT_MIN_ANGLE,
        ]:
            return self._int2coord(res[0])
        elif genre == ProtocolCode.GET_ANGLES_COORDS:
            r = []
            for index in range(len(res)):
                if index < 7:
                    r.append(self._int2angle(res[index]))
                elif index < 10:
                    r.append(self._int2coord(res[index]))
                else:
                    r.append(self._int2angle(res[index]))
            return r
        elif genre == ProtocolCode.GO_ZERO:
            r = []
            if res:
                if 1 not in res[1:]:
                    return res[0]
                else:
                    for i in range(1, len(res)):
                        if res[i] == 1:
                            r.append(i)
            return r
        elif genre in [ProtocolCode.COBOTX_GET_ANGLE, ProtocolCode.COBOTX_GET_SOLUTION_ANGLES,
                       ProtocolCode.GET_POS_OVER]:
            return self._int2angle(res[0])
        elif genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            for i in range(9, len(res)):
                if res[i] != 0:
                    data = bin(res[i])[2:]
                    res[i] = []
                    while len(data) != 16:
                        data = "0" + data
                    for j in range(16):
                        if data[j] != "0":
                            error_id = 15 - j
                            res[i].append(error_id)
                    if not res[i]:
                        res[i] = 0
            return res
        elif genre == ProtocolCode.GET_MOTORS_RUN_ERR:
            for i in range(len(res)):
                if res[i] != 0:
                    data = bin(res[i])[2:]
                    res[i] = []
                    while len(data) != 16:
                        data = "0" + data
                    for j in range(16):
                        if data[j] != "0":
                            error_id = 15 - j
                            res[i].append(error_id)
                    if not res[i]:
                        res[i] = 0
            return res
        else:
            return res

    def _mesg(self, genre, *args, **kwargs):
        read_data = super(Pro630Api, self)._mesg(genre, *args, **kwargs)
        decode_result = self._decode_command(genre, read_data)
        return self._parsing_parameters(genre, decode_result)

    def _joint_limit_init(self):
        for joint, index in enumerate(range(6), start=1):
            if self._joint_max_angles[index] is None:
                angle = self.get_joint_max_angle(joint)
                self._joint_max_angles[index] = angle
            if self._joint_min_angles[index] is None:
                angle = self.get_joint_min_angle(joint)
                self._joint_min_angles[index] = angle

    def _joint_limit_judge(self, angles):
        offset = 3
        try:
            for joint, angle in enumerate(angles, start=1):
                min_angle = self._joint_min_angles[joint - 1]
                max_angle = self._joint_max_angles[joint - 1]
                if min_angle + offset < angle < max_angle - offset:
                    continue

                if self.language == "zh_CN":
                    return f"当前关节{joint}角度为{angle}, 角度范围为: {min_angle} ~ {max_angle}"
                return f"The angle of the current joint {joint} is {angle}, and the angle range is: {min_angle} ~ {max_angle}"
        except TypeError:
            return "joint limit error"
        return "over limit error {}".format(angles)

    def _check_coords(self, new_coords, is_print=False):
        if isinstance(new_coords, list):
            return "check coords error"

        error_info = ""
        first_three = new_coords[:3]
        first_three[2] -= 83.64

        magnitude = np.linalg.norm(first_three)  # Calculate the Euclidean norm (magnitude)
        if is_print == 1:
            if self.language == "zh_CN":
                error_info += f"当前臂展为{magnitude}, 最大的臂展为{self.arm_span_maximum}"
            else:
                error_info += f"Arm span is {magnitude}, max is {self.arm_span_maximum}"

        return error_info

    def _status_explain(self, status):
        error_info = _interpret_status_code(self.language, status)
        if 0x00 < status <= 0x07:
            self._joint_limit_init()
            angles = self.get_angles()
            error_info += self._joint_limit_judge(angles)
        elif status in [32, 33]:
            error_coords = self.get_coords()
            error_info += self._check_coords(error_coords, is_print=True)
        elif status == 36:
            angles = self.get_angles()
            if self.language == "zh_CN":
                error_info += f"当前角度为{angles}, 检测到奇异点"
            else:
                error_info += f"The current angle is {angles}, and a singularity is detected"
        return error_info


class Pro630(Pro630Api):

    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False, save_serial_log=False):
        self._serial_port = setup_serial_connect(port, baudrate, timeout)
        super(Pro630, self).__init__(debug, save_serial_log)
        import RPi.GPIO as GPIO
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup(self.power_control_1, GPIO.IN)
        GPIO.setup(self.power_control_2, GPIO.OUT)

    def open(self):
        self._serial_port.open()

    def close(self):
        self._serial_port.close()

    def power_on(self, delay=2):
        import RPi.GPIO as GPIO
        GPIO.output(self.power_control_2, GPIO.HIGH)
        time.sleep(delay)
        return super(Pro630, self).power_on()

    def power_off(self):
        import RPi.GPIO as GPIO
        res = super(Pro630, self).power_off()
        GPIO.output(self.power_control_2, GPIO.LOW)
        return res

    def power_on_only(self):
        import RPi.GPIO as GPIO
        GPIO.output(self.power_control_2, GPIO.HIGH)

    def set_basic_output(self, pin_no, pin_signal):
        """Set basic output.IO low-level output high-level, high-level output high resistance state

        Args:
            pin_no: pin port number. range 1 ~ 6
            pin_signal: 0 / 1
        """
        import RPi.GPIO as GPIO
        pin_number = self.pin_numbered_mapping_table.get(pin_no, None)
        if pin_number is None:
            raise ValueError("pin_no must be in range 1 ~ 6")
        GPIO.setup(pin_number, GPIO.OUT)
        GPIO.output(pin_number, pin_signal)

    def get_basic_input(self, pin_no):
        """Get basic input.

        Args:
            pin_no: pin port number. range 1 ~ 6

        Return:
            1 - high
            0 - low
        """
        import RPi.GPIO as GPIO
        pin_number = self.pin_numbered_mapping_table.get(pin_no, None)
        if pin_number is None:
            raise ValueError("pin_no must be in range 1 ~ 6")
        GPIO.setup(pin_number, GPIO.IN)
        return GPIO.input(pin_number)

    def send_angles_sync(self, angles, speed):
        angles = [self._angle2int(angle) for angle in angles]
        self.calibration_parameters(class_name=self.__class__.__name__, angles=angles, speed=speed)
        return self._mesg(ProtocolCode.SEND_ANGLES, angles, speed, no_return=True)

    def set_monitor_mode(self, mode):
        raise NotImplementedError("Pro630 does not support monitor mode")

    def get_monitor_mode(self):
        raise NotImplementedError("Pro630 does not support monitor mode")



"""pymycobot package initialization module."""

# ruff: noqa: F401
import datetime
import sys
from pymycobot.mycobot280 import MyCobot280
from pymycobot.mycobot280rdkx5 import MyCobot280RDKX5, MyCobot280RDKX5Socket
from pymycobot.mypalletizer260 import MyPalletizer260
from pymycobot.mecharm270 import MechArm270
from pymycobot.mycobot280socket import MyCobot280Socket
from pymycobot.mycobot320socket import MyCobot320Socket
from pymycobot.mycobot320 import MyCobot320
from pymycobot.generate import CommandGenerator
from pymycobot.Interface import MyBuddyCommandGenerator
from pymycobot.mycobot import MyCobot
from pymycobot.mybuddy import MyBuddy
from pymycobot.mecharm import MechArm
from pymycobot.mypalletizer import MyPalletizer
from pymycobot.mycobotsocket import MyCobotSocket
from pymycobot.genre import Angle, Coord
from pymycobot import utils
from pymycobot.mybuddysocket import MyBuddySocket
from pymycobot.ultraArm import ultraArm
from pymycobot.mybuddybluetooth import MyBuddyBlueTooth
from pymycobot.mypalletizersocket import MyPalletizerSocket
from pymycobot.myarm import MyArm
from pymycobot.myarmsocket import MyArmSocket
from pymycobot.elephantrobot import ElephantRobot
from pymycobot.mercury import Mercury
from pymycobot.myagv import MyAgv
from pymycobot.myagvpro import MyAGVPro
from pymycobot.myagvpro_socket import MyAGVProSocket
from pymycobot.myagvpro_bluetooth import MyAGVProBluetooth
from pymycobot.myarmsocket import MyArmSocket
from pymycobot.mecharmsocket import MechArmSocket

# from pymycobot.mycobotpro630 import Phoenix
from pymycobot.mercurysocket import MercurySocket
from pymycobot.myarmm import MyArmM
from pymycobot.myarmc import MyArmC
from pymycobot.pro630 import Pro630
from pymycobot.pro630client import Pro630Client
from pymycobot.pro400 import Pro400
from pymycobot.pro400client import Pro400Client
from pymycobot.pro450client import Pro450Client
from pymycobot.myarmm_control import MyArmMControl
from pymycobot.mercurychassis_api import ChassisControl
from pymycobot.conveyor_api import ConveyorAPI
from pymycobot.ultraArmP340 import ultraArmP340
from pymycobot.ultraArmP1 import UltraArmP1
from pymycobot.exoskeleton import Exoskeleton
from pymycobot.exoskeletonsocket import ExoskeletonSocket
from pymycobot.mybuddyemoticon import MyBuddyEmoticon


__all__ = [
    "MyPalletizer260",
    "MechArm270",
    "MyCobot280",
    "MyCobot320",
    "MyCobot",
    "CommandGenerator",
    "Angle",
    "Coord",
    "utils",
    "MyPalletizer",
    "MyCobotSocket",
    "MyBuddyCommandGenerator",
    "MyBuddy",
    "MyBuddySocket",
    "MyBuddyBlueTooth",
    "ultraArm",
    "MyPalletizerSocket",
    "MechArm",
    "MyArm",
    "ElephantRobot",
    "Mercury",
    "MyAgv",
    "MyAGVPro",
    "MyAGVProSocket",
    "MyAGVProBluetooth",
    "MechArmSocket",
    "MyArmSocket",
    "MercurySocket",
    "MyArmM",
    "MyArmC",
    # "Phoenix",
    "Pro630",
    "Pro630Client",
    "Pro400",
    "Pro400Client",
    "MyCobot280Socket",
    "MyCobot320Socket",
    "MyArmMControl",
    "ChassisControl",
    "ConveyorAPI",
    "MyCobot280RDKX5",
    "MyCobot280RDKX5Socket",
    "ultraArmP340",
    "Exoskeleton",
    "ExoskeletonSocket",
    "MyBuddyEmoticon",
    "Pro450Client",
    "UltraArmP1",
]

__version__ = "4.0.3"
__author__ = "Elephantrobotics"
__email__ = "weiquan.xu@elephantrobotics.com"
__git_url__ = "https://github.com/elephantrobotics/pymycobot"
__copyright__ = f"CopyRight (c) 2020-{datetime.datetime.now().year} Shenzhen Elephantrobotics technology"

# For raspberry mycobot 280.
PI_PORT = "/dev/ttyAMA0"
PI_BAUD = 1000000
